#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Device class for controlling a Schneider Electric ILS2T stepper drive over modbus TCP.
"""

import logging
import struct
import sys
from datetime import timedelta
from enum import Flag, IntEnum
from time import sleep, time
from typing import Any, ClassVar, cast

import aenum
from bitstring import BitArray
from pymodbus.client import ModbusTcpClient

from hvl_ccb.comm.modbus_tcp import (
    ModbusTcpCommunication,
    ModbusTcpCommunicationConfig,
    ModbusTcpConnectionFailedError,
)
from hvl_ccb.configuration import configdataclass
from hvl_ccb.dev.base import DeviceError, SingleCommDevice
from hvl_ccb.utils.typing import Number

logger = logging.getLogger(__name__)


class ILS2TError(DeviceError):
    """
    Error to indicate problems with the SE ILS2T stepper motor.
    """


class IoScanningModeValueError(ILS2TError):
    """
    Erorr to indicate that the selected IO scanning mode is invalid.
    """


class ScalingFactorValueError(ILS2TError):
    """
    Error to indicate that a scaling factor value is invalid.
    """


@configdataclass
class ILS2TModbusTcpCommunicationConfig(ModbusTcpCommunicationConfig):
    """
    Configuration dataclass for Modbus/TCP communciation specific for the Schneider
    Electric ILS2T stepper motor.
    """

    #: The unit has to be 255 such that IO scanning mode works.
    unit: int = 255


class ILS2TModbusTcpCommunication(ModbusTcpCommunication):
    """
    Specific implementation of Modbus/TCP for the Schneider Electric ILS2T stepper
    motor.
    """

    @staticmethod
    def config_cls():
        return ILS2TModbusTcpCommunicationConfig


@configdataclass
class ILS2TConfig:
    """
    Configuration for the ILS2T stepper motor device.
    """

    #: initial maximum RPM for the motor, can be set up to 3000 RPM. The user is
    #: allowed to set a new max RPM at runtime using :meth:`ILS2T.set_max_rpm`,
    #: but the value must never exceed this configuration setting.
    rpm_max_init: int = 1500
    wait_sec_post_enable: Number = 1
    wait_sec_max_disable: Number = 10
    wait_sec_post_cannot_disable: Number = 1
    wait_sec_post_relative_step: Number = 2
    wait_sec_post_absolute_position: Number = 2

    def clean_values(self) -> None:
        if not 0 < self.rpm_max_init <= 3000:
            msg = "Maximum RPM for the motor must be integer number between 1 and 3000."
            raise ValueError(msg)
        if self.wait_sec_post_enable <= 0:
            msg = "Wait time post motor enabling must be a positive value (in seconds)."
            raise ValueError(msg)
        if self.wait_sec_max_disable < 0:
            msg = (
                "Maximal wait time for attempting to disable motor must be a "
                "non-negative value (in seconds)."
            )
            raise ValueError(msg)
        if self.wait_sec_post_cannot_disable <= 0:
            msg = (
                "Wait time post failed motor disable attempt must be a positive value "
                "(in seconds)."
            )
            raise ValueError(msg)
        if self.wait_sec_post_relative_step <= 0:
            msg = (
                "Wait time post motor relative step must be a positive value "
                "(in seconds)."
            )
            raise ValueError(msg)
        if self.wait_sec_post_absolute_position <= 0:
            msg = (
                "Wait time post motor absolute position change must be a positive "
                "value (in seconds)."
            )
            raise ValueError(msg)


class ILS2TRegDatatype(aenum.Enum, init="min max"):  # type: ignore[call-arg]
    """
    Modbus Register Datatypes for Schneider Electric ILS2T stepper drive.

    From the manual of the drive:

    =========== =========== ============== =============
    datatype    byte        min            max
    =========== =========== ============== =============
    INT8        1 Byte      -128           127
    UINT8       1 Byte      0              255
    INT16       2 Byte      -32_768        32_767
    UINT16      2 Byte      0              65_535
    INT32       4 Byte      -2_147_483_648 2_147_483_647
    UINT32      4 Byte      0              4_294_967_295
    BITS        just 32bits N/A            N/A
    =========== =========== ============== =============

    """

    INT32 = -2_147_483_648, 2_147_483_647

    def is_in_range(self, value: int) -> bool:
        return self.min <= value <= self.max


class ILS2TRegAddr(IntEnum):
    """
    Modbus Register Adresses for for Schneider Electric ILS2T stepper drive.
    """

    POSITION = 7706  # INT32 position of the motor in user defined units
    IO_SCANNING = 6922  # BITS start register for IO scanning control
    # and status
    TEMP = 7200  # INT16 temperature of motor
    VOLT = 7198  # UINT16 dc voltage of motor
    SCALE = 1550  # INT32 user defined steps per revolution
    ACCESS_ENABLE = 282  # BITS not documented register
    # to enable access via IO scanning
    JOGN_FAST = 10506  # UINT16 revolutions per minute for fast Jog (1 to 3000)
    JOGN_SLOW = 10504  # UINT16 revolutions per minute
    # for slow Jog (1 to 3000)

    RAMP_TYPE = 1574  # INT16 ramp type, 0: linear / -1: motor optimized
    RAMP_ACC = 1556  # UINT32 acceleration
    RAMP_DECEL = 1558  # UINT32 deceleration
    RAMP_N_MAX = 1554  # UINT16 max rpm
    FLT_INFO = 15362  # 22 registers, code for error
    FLT_MEM_RESET = 15114  # UINT16 reset fault memory
    FLT_MEM_DEL = 15112  # UINT16 delete fault memory


class ILS2T(SingleCommDevice):
    """
    Schneider Electric ILS2T stepper drive class.
    """

    RegDatatype = ILS2TRegDatatype
    """Modbus Register Datatypes
    """
    RegAddr = ILS2TRegAddr
    """Modbus Register Adresses
    """

    class Mode(IntEnum):
        """
        ILS2T device modes
        """

        PTP = 3  # point to point
        JOG = 1

    class ActionsPtp(IntEnum):
        """
        Allowed actions in the point to point mode (`ILS2T.Mode.PTP`).
        """

        ABSOLUTE_POSITION = 0
        RELATIVE_POSITION_TARGET = 1
        RELATIVE_POSITION_MOTOR = 2

    ACTION_JOG_VALUE = 0
    """
    The single action value for `ILS2T.Mode.JOG`
    """

    # Note: don't use IntFlag here - it allows other then enumerated values
    # The behaviour changed with py311
    if sys.version_info < (3, 11):

        class Ref16Jog(Flag):
            """
            Allowed values for ILS2T ref_16 register (the shown values are the integer
            representation of the bits), all in Jog mode = 1
            """

            NONE = 0
            POS = 1
            NEG = 2
            FAST = 4
            # allowed combinations
            POS_FAST = POS | FAST
            NEG_FAST = NEG | FAST
    else:
        from enum import STRICT

        class Ref16Jog(Flag, boundary=STRICT):
            """
            Allowed values for ILS2T ref_16 register (the shown values are the integer
            representation of the bits), all in Jog mode = 1
            """

            NONE = 0
            POS = 1
            NEG = 2
            FAST = 4
            # allowed combinations
            POS_FAST = POS | FAST
            NEG_FAST = NEG | FAST

    class State(IntEnum):
        """
        State machine status values
        """

        QUICKSTOP = 7
        READY = 4
        ON = 6

    DEFAULT_IO_SCANNING_CONTROL_VALUES: ClassVar[dict[str, Any]] = {
        "action": ActionsPtp.RELATIVE_POSITION_MOTOR.value,
        "mode": Mode.PTP.value,
        "disable_driver_di": 0,
        "enable_driver_en": 0,
        "quick_stop_qs": 0,
        "fault_reset_fr": 0,
        "execute_stop_sh": 0,
        "reset_stop_ch": 0,
        "continue_after_stop_cu": 0,
        "ref_16": ILS2TConfig.rpm_max_init,
        "ref_32": 0,
    }
    """
    Default IO Scanning control mode values
    """

    def __init__(self, com, dev_config=None) -> None:
        """
        Constructor for ILS2T.

        :param com: object to use as communication protocol.
        """

        # Call superclass constructor
        super().__init__(com, dev_config)

        # toggle reminder bit
        self._mode_toggle_mt = 0

        self.flt_list: list[dict[int, dict[str, Any]]] = []

    @staticmethod
    def default_com_cls():
        return ILS2TModbusTcpCommunication

    @staticmethod
    def config_cls():
        return ILS2TConfig

    def start(self) -> None:
        """
        Start this device.
        """

        logger.info(f"Starting device {self}")
        try:
            # try opening the port
            super().start()
        except ModbusTcpConnectionFailedError:
            logger.exception("Error with ModbusTCP Connection")
            raise

        # writing 1 to register ACCESS_ENABLE allows to use the IO scanning mode.
        #  This is not documented in the manual!
        self.com.write_registers(self.RegAddr.ACCESS_ENABLE.value, [0, 1])

        # set maximum RPM from init config
        self.set_max_rpm(self.config.rpm_max_init)

    def stop(self) -> None:
        """
        Stop this device. Disables the motor (applies brake), disables access and
        closes the communication protocol.
        """

        logger.info(f"Stopping device {self}")
        self.disable()
        self.com.write_registers(self.RegAddr.ACCESS_ENABLE.value, [0, 0])
        super().stop()

    def get_status(self) -> dict[str, int | list[bool]]:
        """
        Perform an IO Scanning read and return the status of the motor.

        :return: dict with status information.
        """

        registers = self.com.read_holding_registers(self.RegAddr.IO_SCANNING.value, 8)
        return self._decode_status_registers(registers)

    def do_ioscanning_write(self, **kwargs: int) -> None:
        """
        Perform a write operation using IO Scanning mode.

        :param kwargs:
            Keyword-argument list with options to send, remaining are taken
            from the defaults.
        """

        self._toggle()
        values = self._generate_control_registers(**kwargs)
        self.com.write_registers(self.RegAddr.IO_SCANNING.value, values)

    def _generate_control_registers(self, **kwargs: int) -> list[int]:
        """
        Generates the control registers for the IO scanning mode.
        It is necessary to write all 64 bit at the same time, so a list of 4 registers
        is generated.

        :param kwargs: Keyword-argument list with options different than the defaults.
        :return: List of registers for the IO scanning mode.
        """

        cleaned_io_scanning_mode = self._clean_ioscanning_mode_values(kwargs)

        action_bits = f"{cleaned_io_scanning_mode['action']:03b}"
        mode_bits = f"{cleaned_io_scanning_mode['mode']:04b}"
        builder = []

        builder.extend(
            ModbusTcpClient.convert_to_registers(
                value=[
                    # add the first byte: Drive control
                    cleaned_io_scanning_mode["disable_driver_di"],
                    cleaned_io_scanning_mode["enable_driver_en"],
                    cleaned_io_scanning_mode["quick_stop_qs"],
                    cleaned_io_scanning_mode["fault_reset_fr"],
                    0,  # has to be 0 per default, no meaning
                    cleaned_io_scanning_mode["execute_stop_sh"],
                    cleaned_io_scanning_mode["reset_stop_ch"],
                    cleaned_io_scanning_mode["continue_after_stop_cu"],
                    # add the second byte: Mode control
                    int(mode_bits[3]),
                    int(mode_bits[2]),
                    int(mode_bits[1]),
                    int(mode_bits[0]),
                    int(action_bits[2]),
                    int(action_bits[1]),
                    int(action_bits[0]),
                    self._mode_toggle_mt,
                ],
                data_type=ModbusTcpClient.DATATYPE.BITS,
                word_order="big",
            )
        )

        # add the third and fourth byte:
        # Ref_16 (either JOG direction/speed, or RPM...)
        builder.extend(
            ModbusTcpClient.convert_to_registers(
                value=cleaned_io_scanning_mode["ref_16"],
                data_type=ModbusTcpClient.DATATYPE.UINT16,
                word_order="big",
            )
        )

        # add 4 bytes Ref_32, Target position
        builder.extend(
            ModbusTcpClient.convert_to_registers(
                value=cleaned_io_scanning_mode["ref_32"],
                data_type=ModbusTcpClient.DATATYPE.INT32,
                word_order="big",
            )
        )
        return builder

    def _clean_ioscanning_mode_values(
        self, io_scanning_values: dict[str, int]
    ) -> dict[str, int]:
        """
        Checks if the constructed mode is valid.

        :param io_scanning_values: Dictionary with register values to check
        :return: Dictionary with cleaned register values
        :raises ValueError: if `io_scanning_values` has unrecognized keys
        :raises IoScanningModeValueError: if either `'mode'` or either of corresponding
            `'action'`, `'ref_16'`, or `'ref_32'` keys of `io_scanning_values` has
            an invalid value.
        """

        # check if there are too much keys that are not recognized
        io_scanning_keys = set(io_scanning_values.keys())
        all_keys = set(self.DEFAULT_IO_SCANNING_CONTROL_VALUES.keys())
        superfluous_keys = io_scanning_keys.difference(all_keys)
        if superfluous_keys:
            msg = f"Unrecognized mode keys: {list(superfluous_keys)}"
            raise ValueError(msg)

        # fill up io_scanning_values with defaults, if they are not set
        for mode_key, default_value in self.DEFAULT_IO_SCANNING_CONTROL_VALUES.items():
            if mode_key not in io_scanning_values:
                io_scanning_values[mode_key] = cast("int", default_value)

        # perform checks depending on mode
        # JOG mode
        if io_scanning_values["mode"] == self.Mode.JOG:
            io_scanning_value = io_scanning_values["action"]
            if io_scanning_value != self.ACTION_JOG_VALUE:
                msg = f"Wrong action: {io_scanning_value}"
                raise IoScanningModeValueError(msg)

            io_scanning_value = io_scanning_values["ref_16"]
            try:
                self.Ref16Jog(io_scanning_value)
            except ValueError as exc:
                msg = f"Wrong value in ref_16 ({io_scanning_value})"
                logger.exception(msg)
                raise IoScanningModeValueError(msg) from exc

            io_scanning_value = io_scanning_values["ref_32"]
            if io_scanning_value != 0:
                msg = f"Wrong value in ref_32 ({io_scanning_value})"
                raise IoScanningModeValueError(msg)

            return io_scanning_values

        # PTP mode
        if io_scanning_values["mode"] == self.Mode.PTP:
            io_scanning_value = io_scanning_values["action"]
            try:
                self.ActionsPtp(io_scanning_value)
            except ValueError as exc:
                msg = f"Wrong action: {io_scanning_value}"
                logger.exception(msg)
                raise IoScanningModeValueError(msg) from exc

            io_scanning_value = io_scanning_values["ref_16"]
            if not self._is_valid_rpm(io_scanning_value):
                msg = f"Wrong value in ref_16 ({io_scanning_value})"
                raise IoScanningModeValueError(msg)

            io_scanning_value = io_scanning_values["ref_32"]
            if not self._is_int32(io_scanning_value):
                msg = f"Wrong value in ref_32 ({io_scanning_value})"
                raise IoScanningModeValueError(msg)

            return io_scanning_values

        # default
        msg = f"Wrong mode: {io_scanning_values['mode']}"
        raise IoScanningModeValueError(msg)

    def _is_valid_rpm(self, num: int) -> bool:
        """
        Checks whether `num` is a valid RPM value.

        :param num: RPM value to check
        :return: `True` if `num` is a valid RPM value, `False` otherwise
        """

        return isinstance(num, int) and 0 < num <= self.config.rpm_max_init

    @classmethod
    def _is_int32(cls, num: int) -> bool:
        """
        Checks whether a number fits in a signed 32-bit integer.

        :param num: is the number to check.
        :return: check result.
        """
        return isinstance(num, int) and cast(
            "ILS2TRegDatatype", ILS2TRegDatatype.INT32
        ).is_in_range(num)

    @staticmethod
    def _decode_status_registers(registers: list[int]) -> dict[str, int | list[bool]]:
        """
        Decodes the the status of the stepper drive, derived from IOscanning.
        register 1 is decoded into "ref_16" (16-bit int)
        register 2 and 3 together into "ref_32" (32-bit int)
        the others are decoded into bit lists

        :param registers: List of 8 registers (6922-6930)
        :return: dict
        """
        decoded = {}

        bit_list_registers = {
            0: ("drive_control", "mode_control"),
            4: ("drive_status_1", "drive_status_2"),
            5: ("mode_status", "drive_input"),
            6: ("action_word_1", "action_word_2"),
            7: ("special_function_1", "special_function_2"),
        }

        for idx, keys in bit_list_registers.items():
            reg_bytes = registers[idx].to_bytes(2, byteorder="big")
            bit_lists = [
                [bool((byte >> bit_position) & 1) for bit_position in range(8)]
                for byte in reg_bytes
            ]
            decoded[keys[0]], decoded[keys[1]] = bit_lists

        # > means big-endian and h signed short int 2 bytes
        decoded["ref_16"] = struct.unpack(">h", struct.pack(">h", registers[1]))[0]

        # Combine register 2 and 3 into a 32-bit integer
        # HH: two unsigned short int 2 bytes, I: big-endian unsigned int 4 bytes
        byte_data = struct.pack(">HH", registers[2], registers[3])
        decoded["ref_32"] = struct.unpack(">I", byte_data)[0]

        return {
            "mode": BitArray(decoded["mode_status"][3::-1]).int,
            "action": BitArray(decoded["mode_control"][6:3:-1]).int,
            "ref_16": decoded["ref_16"],
            "ref_32": decoded["ref_32"],
            "state": BitArray(decoded["drive_status_2"][3::-1]).int,
            "fault": decoded["drive_status_2"][6],
            "warn": decoded["drive_status_2"][7],
            "halt": decoded["drive_status_1"][0],
            "motion_zero": decoded["action_word_2"][6],
            "turning_positive": decoded["action_word_2"][7],
            "turning_negative": decoded["action_word_1"][0],
        }

    def _toggle(self) -> None:
        """
        To activate a command it is necessary to toggle the MT bit first.
        """

        self._mode_toggle_mt = 0 if self._mode_toggle_mt else 1

    def write_relative_step(self, steps: int) -> None:
        """
        Write instruction to turn the motor the relative amount of steps. This function
        does not enable or disable the motor automatically.

        :param steps: Number of steps to turn the motor.
        """
        max_step = self.RegDatatype.INT32.max  # type: ignore[attr-defined]
        # use _is_int32 instead?
        if not abs(steps) < max_step:
            logger.warning(f"number of steps is too big: {steps}")

        logger.info(f"Perform number of steps: {steps}")

        self.do_ioscanning_write(
            enable_driver_en=1,
            mode=self.Mode.PTP.value,
            action=self.ActionsPtp.RELATIVE_POSITION_MOTOR.value,
            ref_32=steps,
        )

    def write_absolute_position(self, position: int) -> None:
        """
        Write instruction to turn the motor until it reaches the absolute position.
        This function does not enable or disable the motor automatically.

        :param position: absolute position of motor in user defined steps.
        """

        max_position = self.RegDatatype.INT32.max  # type: ignore[attr-defined]
        # use _is_int32 instead?
        if not abs(position) < max_position:
            logger.warning(f"position is out of range: {position}")

        logger.info(f"Absolute position: {position}")

        self.do_ioscanning_write(
            enable_driver_en=1,
            mode=self.Mode.PTP.value,
            action=self.ActionsPtp.ABSOLUTE_POSITION.value,
            ref_32=position,
        )

    def _is_position_as_expected(
        self, position_expected: int, position_actual: int, err_msg: str
    ) -> bool:
        """
        Check if actual drive position is a expected. If expectation is not met,
        check for possible drive error and log the given error message with appropriate
        level of severity. Do not raise error; instead, return `bool` stating if
        expectation was met.

        :param position_expected: Expected drive position.
        :param position_actual: Actual drive position.
        :param err_msg: Error message to log if expectation is not met.
        :return: `True` if actual position is as expected, `False` otherwise.
        """
        as_expected = position_expected == position_actual
        if not as_expected:
            flt_dict = self.get_error_code()
            self.flt_list.append(flt_dict)
            if "empty" in flt_dict[0]:
                logger.warning(
                    "no error in drive, something different must have gone wrong"
                )
                logger.warning(err_msg)
            else:
                logger.critical("error in drive, drive is know maybe locked")
                logger.critical(err_msg)
        return as_expected

    def execute_relative_step(self, steps: int) -> bool:
        """
        Execute a relative step, i.e. enable motor, perform relative steps,
        wait until done and disable motor afterwards.

        Check position at the end if wrong do not raise error; instead just log and
        return check result.

        :param steps: Number of steps.
        :return: `True` if actual position is as expected, `False` otherwise.
        """
        logger.info(f"Motor steps requested: {steps}")

        with self.com.access_lock:
            position_before = self.get_position()

            self.enable()
            sleep(self.config.wait_sec_post_enable)
            self.write_relative_step(steps)
            sleep(self.config.wait_sec_post_relative_step)
            self.disable(log_warn=False)

            # check if steps were made
            position_after = self.get_position()
            return self._is_position_as_expected(
                position_before + steps,
                position_after,
                "The position does not align with the requested step number. "
                f"Before: {position_before}, after: {position_after}, "
                f"requested: {steps}, "
                f"real difference: {position_after - position_before}.",
            )

    def execute_absolute_position(self, position: int) -> bool:
        """
        Execute a absolute position change, i.e. enable motor, perform absolute
        position change, wait until done and disable motor afterwards.

        Check position at the end if wrong do not raise error; instead just log and
        return check result.

        :param position: absolute position of motor in user defined steps.
        :return: `True` if actual position is as expected, `False` otherwise.
        """
        logger.info(f"absolute position requested: {position}")

        with self.com.access_lock:
            position_before = self.get_position()

            self.enable()
            sleep(self.config.wait_sec_post_enable)
            self.write_absolute_position(position)
            sleep(self.config.wait_sec_post_absolute_position)
            self.disable(log_warn=False)

            # check if steps were made
            position_after = self.get_position()
            return self._is_position_as_expected(
                position,
                position_after,
                "The position does not align with the requested absolute position."
                f"Before: {position_before}, after: {position_after}, "
                f"requested: {position}.",
            )

    def disable(
        self,
        log_warn: bool = True,
        wait_sec_max: int | None = None,
    ) -> bool:
        """
        Disable the driver of the stepper motor and enable the brake.

        Note: the driver cannot be disabled if the motor is still running.

        :param log_warn: if log a warning in case the motor cannot be disabled.
        :param wait_sec_max: maximal wait time for the motor to stop running and to
            disable it; by default, with `None`, use a config value
        :return: `True` if disable request could and was sent, `False` otherwise.
        """
        if wait_sec_max is None:
            wait_sec_max = self.config.wait_sec_max_disable

        try_disable = True
        elapsed_time = 0.0
        start_time = time()
        while try_disable:
            can_disable = bool(self.get_status()["motion_zero"])
            if can_disable:
                logger.info("Disable motor, brake.")
                self.do_ioscanning_write(enable_driver_en=0, disable_driver_di=1)
            elif log_warn:
                logger.warning("Cannot disable motor, still running!")
            elapsed_time += time() - start_time

            try_disable = not can_disable and elapsed_time < wait_sec_max
            if try_disable:
                sleep(self.config.wait_sec_post_cannot_disable)

        return can_disable

    def enable(self) -> None:
        """
        Enable the driver of the stepper motor and disable the brake.
        """

        self.do_ioscanning_write(enable_driver_en=1, disable_driver_di=0)
        logger.info("Enable motor, disable brake.")

    def get_position(self) -> int:
        """
        Read the position of the drive and store into status.

        :return: Position step value
        """

        value = self.com.read_input_registers(self.RegAddr.POSITION.value, 2)
        return self._decode_32bit(value, True)

    def get_temperature(self) -> int:
        """
        Read the temperature of the motor.

        :return: Temperature in degrees Celsius.
        """

        value = self.com.read_input_registers(self.RegAddr.TEMP.value, 2)
        return self._decode_32bit(value, True)

    def get_dc_volt(self) -> float:
        """
        Read the DC supply voltage of the motor.

        :return: DC input voltage.
        """

        value = self.com.read_input_registers(self.RegAddr.VOLT.value, 2)
        return self._decode_32bit(value, True) / 10

    @staticmethod
    def _decode_32bit(registers: list[int], signed: bool = True) -> int:
        """
        Decode two 16-bit ModBus registers to a 32-bit integer.

        :param registers: list of two register values
        :param signed: True, if register containes a signed value
        :return: integer representation of the 32-bit register
        """

        if signed:
            return ModbusTcpClient.convert_from_registers(
                registers=registers, data_type=ModbusTcpClient.DATATYPE.INT32
            )

        return ModbusTcpClient.convert_from_registers(
            registers=registers, data_type=ModbusTcpClient.DATATYPE.UINT32
        )

    def user_steps(self, steps: int = 16384, revolutions: int = 1) -> None:
        """
        Define steps per revolution.
        Default is 16384 steps per revolution.
        Maximum precision is 32768 steps per revolution.

        :param steps: number of steps in `revolutions`.
        :param revolutions: number of revolutions corresponding to `steps`.
        """

        if not self._is_int32(revolutions):
            err_msg = f"Wrong scaling factor: revolutions = {revolutions}"
            logger.error(err_msg)
            raise ScalingFactorValueError(err_msg)

        if not self._is_int32(steps):
            err_msg = f"Wrong scaling factor: steps = {steps}"
            logger.error(err_msg)
            raise ScalingFactorValueError(err_msg)

        builder = []
        builder.extend(
            ModbusTcpClient.convert_to_registers(
                value=steps,
                data_type=ModbusTcpClient.DATATYPE.INT32,
                word_order="big",
            )
        )
        builder.extend(
            ModbusTcpClient.convert_to_registers(
                value=revolutions,
                data_type=ModbusTcpClient.DATATYPE.INT32,
                word_order="big",
            )
        )
        self.com.write_registers(self.RegAddr.SCALE.value, builder)

    def quickstop(self) -> None:
        """
        Stops the motor with high deceleration rate and falls into error state. Reset
        with `reset_error` to recover into normal state.
        """

        logger.warning("Motor QUICK STOP.")
        self.do_ioscanning_write(quick_stop_qs=1)

    def reset_error(self) -> None:
        """
        Resets the motor into normal state after quick stop or another error occured.
        """

        logger.info("Reset motor after fault or quick stop.")
        self.do_ioscanning_write(fault_reset_fr=1)

    def jog_run(self, direction: bool = True, fast: bool = False) -> None:
        """
        Slowly turn the motor in positive direction.
        """

        status = self.get_status()

        if status["mode"] != self.Mode.JOG and not status["motion_zero"]:
            logger.error("Motor is not in Jog mode or standstill, abort.")
            return

        if status["state"] != self.State.ON:
            # need to enable first
            logger.error("Motor is not enabled or in error state. Try .enable()")
            return

        ref_16 = self.Ref16Jog.NONE

        if direction:
            ref_16 = ref_16 | self.Ref16Jog.POS
            logger.info("Jog mode in positive direction enabled.")
        else:
            ref_16 = ref_16 | self.Ref16Jog.NEG
            logger.info("Jog mode in negative direction enabled.")

        if fast:
            ref_16 = ref_16 | self.Ref16Jog.FAST

        self.do_ioscanning_write(
            mode=self.Mode.JOG.value,
            action=self.ACTION_JOG_VALUE,
            enable_driver_en=1,
            ref_16=ref_16.value,
        )

    def jog_stop(self) -> None:
        """
        Stop turning the motor in Jog mode.
        """

        logger.info("Stop in Jog mode.")

        self.do_ioscanning_write(
            mode=self.Mode.JOG.value,
            action=self.ACTION_JOG_VALUE,
            enable_driver_en=1,
            ref_16=0,
        )

    def set_jog_speed(self, slow: int = 60, fast: int = 180) -> None:
        """
        Set the speed for jog mode. Default values correspond to startup values of
        the motor.

        :param slow: RPM for slow jog mode.
        :param fast: RPM for fast jog mode.
        """

        logger.info(f"Setting Jog RPM. Slow = {slow} RPM, Fast = {fast} RPM.")
        self.com.write_registers(self.RegAddr.JOGN_SLOW.value, [0, slow])
        self.com.write_registers(self.RegAddr.JOGN_FAST.value, [0, fast])

    def get_error_code(self) -> dict[int, dict[str, Any]]:
        """
        Read all messages in fault memory.
        Will read the full error message and return the decoded values.
        At the end the fault memory of the motor will be deleted.
        In addition, reset_error is called to re-enable the motor for operation.

        :return: Dictionary with all information
        """

        ret_dict = {}
        self.com.write_registers(self.RegAddr.FLT_MEM_RESET.value, [0, 1])
        for i in range(10):
            registers = self.com.read_input_registers(self.RegAddr.FLT_INFO.value, 22)
            decoded = {
                "ignored0": None,
                "error_code": ModbusTcpClient.convert_from_registers(
                    registers=registers[1:2],
                    data_type=ModbusTcpClient.DATATYPE.UINT16,
                    word_order="big",
                ),
                "ignored1": None,
                "error_class": ModbusTcpClient.convert_from_registers(
                    registers=registers[3:4],
                    data_type=ModbusTcpClient.DATATYPE.UINT16,
                    word_order="big",
                ),
                "error_time": ModbusTcpClient.convert_from_registers(
                    registers=registers[4:6],
                    data_type=ModbusTcpClient.DATATYPE.UINT32,
                    word_order="big",
                ),
                "ignored2": None,
                "error_addition": ModbusTcpClient.convert_from_registers(
                    registers=registers[7:8],
                    data_type=ModbusTcpClient.DATATYPE.UINT16,
                    word_order="big",
                ),
                "ignored3": None,
                "error_no_cycle": ModbusTcpClient.convert_from_registers(
                    registers=registers[9:10],
                    data_type=ModbusTcpClient.DATATYPE.UINT16,
                    word_order="big",
                ),
                "ignored4": None,
                "error_after_enable": ModbusTcpClient.convert_from_registers(
                    registers=registers[11:12],
                    data_type=ModbusTcpClient.DATATYPE.UINT16,
                    word_order="big",
                ),
                "ignored5": None,
                "error_voltage_dc": ModbusTcpClient.convert_from_registers(
                    registers=registers[13:14],
                    data_type=ModbusTcpClient.DATATYPE.UINT16,
                    word_order="big",
                ),
                "ignored6": None,
                "error_rpm": ModbusTcpClient.convert_from_registers(
                    registers=registers[15:16],
                    data_type=ModbusTcpClient.DATATYPE.INT16,
                    word_order="big",
                ),
                "ignored7": None,
                "error_current": ModbusTcpClient.convert_from_registers(
                    registers=registers[17:18],
                    data_type=ModbusTcpClient.DATATYPE.UINT16,
                    word_order="big",
                ),
                "ignored8": None,
                "error_amplifier_temperature": ModbusTcpClient.convert_from_registers(
                    registers=registers[19:20],
                    data_type=ModbusTcpClient.DATATYPE.INT16,
                    word_order="big",
                ),
                "ignored9": None,
                "error_device_temperature": ModbusTcpClient.convert_from_registers(
                    registers=registers[21:22],
                    data_type=ModbusTcpClient.DATATYPE.INT16,
                    word_order="big",
                ),
            }
            flt_dict = {
                "error_code": hex(decoded["error_code"])[2:],
                "error_class": decoded["error_class"],
                "error_time": timedelta(seconds=decoded["error_time"]),
                "error_addition": decoded["error_addition"],
                "error_no_cycle": decoded["error_no_cycle"],
                "error_after_enable": timedelta(seconds=decoded["error_after_enable"]),
                "error_voltage_dc": decoded["error_voltage_dc"] / 10,
                "error_rpm": decoded["error_rpm"],
                "error_current": decoded["error_current"] / 100,
                "error_amplifier_temperature": decoded["error_amplifier_temperature"],
                "error_device_temperature": decoded["error_device_temperature"],
            }
            ret_dict[i] = flt_dict
            if flt_dict["error_code"] == "0":
                flt_dict = {"empty": None}
                ret_dict = {i: flt_dict}
                break
        self.com.write_registers(self.RegAddr.FLT_MEM_DEL.value, [0, 1])
        self.reset_error()
        return ret_dict

    def set_max_rpm(self, rpm: int) -> None:
        """
        Set the maximum RPM.

        :param rpm: revolution per minute ( 0 < rpm <= RPM_MAX)
        :raises ILS2TError: if RPM is out of range
        """

        if self._is_valid_rpm(rpm):
            self.DEFAULT_IO_SCANNING_CONTROL_VALUES["ref_16"] = rpm
            self.com.write_registers(self.RegAddr.RAMP_N_MAX.value, [0, rpm])
        else:
            msg = f"RPM out of range: {rpm} not in (0, {self.config.rpm_max_init}]"
            raise ILS2TError(msg)

    def set_ramp_type(self, ramp_type: int = -1) -> None:
        """
        Set the ramp type. There are two options available:
            0:  linear ramp
            -1: motor optimized ramp

        :param ramp_type: 0: linear ramp | -1: motor optimized ramp
        """

        self.com.write_registers(self.RegAddr.RAMP_TYPE.value, [0, ramp_type])

    def set_max_acceleration(self, rpm_minute: int) -> None:
        """
        Set the maximum acceleration of the motor.

        :param rpm_minute: revolution per minute per minute
        """

        builder = []
        builder.extend(
            ModbusTcpClient.convert_to_registers(
                value=rpm_minute,
                data_type=ModbusTcpClient.DATATYPE.UINT32,
                word_order="big",
            )
        )
        self.com.write_registers(self.RegAddr.RAMP_ACC.value, builder)

    def set_max_deceleration(self, rpm_minute: int) -> None:
        """
        Set the maximum deceleration of the motor.

        :param rpm_minute: revolution per minute per minute
        """

        builder = []
        builder.extend(
            ModbusTcpClient.convert_to_registers(
                value=rpm_minute,
                data_type=ModbusTcpClient.DATATYPE.UINT32,
                word_order="big",
            )
        )
        self.com.write_registers(self.RegAddr.RAMP_DECEL.value, builder)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for the .comm.labjack_ljm subpackage.
"""

import logging
from collections import abc

import pytest

from hvl_ccb.comm.labjack_ljm import (
    LJMCommunication,
    LJMCommunicationConfig,
    LJMCommunicationError,
)

logging.disable(logging.ERROR)


@pytest.fixture(scope="module")
def testconfig():
    return {
        # identifier = -2 specifies LJM DEMO mode, see
        # https://labjack.com/support/software/api/ljm/demo-mode
        # however `ljm.getHandleInfo(-2)` raises
        # `LJMError: LJM library error code 1224 LJME_DEVICE_NOT_OPEN`
        "identifier": "-2"
    }


@pytest.fixture
def com(testconfig):
    with LJMCommunication(testconfig) as com:
        yield com


def test_labjack_config_cleaning() -> None:
    with pytest.raises(ValueError):
        LJMCommunicationConfig(device_type="T5")

    with pytest.raises(ValueError):
        LJMCommunicationConfig(device_type="T7-Pro")
    config = LJMCommunicationConfig(device_type="T7_PRO")
    assert config.device_type is LJMCommunicationConfig.DeviceType.T7_PRO

    with pytest.raises(ValueError):
        LJMCommunicationConfig(connection_type="Gardena")
    config = LJMCommunicationConfig(connection_type="TCP")
    assert config.connection_type is LJMCommunicationConfig.ConnectionType.TCP


def test_labjack_config_enums() -> None:
    config = LJMCommunicationConfig(device_type=LJMCommunicationConfig.DeviceType.T4)
    assert config.device_type == "T4"
    device_type = LJMCommunicationConfig.DeviceType.get_by_p_id(config.device_type.p_id)
    assert device_type is LJMCommunicationConfig.DeviceType.T4
    # test device with unambiguous Product ID
    config = LJMCommunicationConfig(
        device_type=LJMCommunicationConfig.DeviceType.T7_PRO
    )
    assert config.device_type == "T7_PRO"
    device_types = LJMCommunicationConfig.DeviceType.get_by_p_id(
        config.device_type.p_id
    )
    assert device_types == [
        LJMCommunicationConfig.DeviceType.T7,
        LJMCommunicationConfig.DeviceType.T7_PRO,
    ]

    config = LJMCommunicationConfig(
        connection_type=LJMCommunicationConfig.ConnectionType.TCP
    )
    assert config.connection_type == "TCP"


def test_open(testconfig) -> None:
    com = LJMCommunication(testconfig)
    assert not com.is_open
    com.open()
    assert com.is_open
    com.open()
    assert com.is_open
    com.close()
    assert not com.is_open

    failing_config = dict(testconfig)
    failing_config["identifier"] = "this_is_an_invalid_identifier"
    com = LJMCommunication(failing_config)
    assert not com.is_open
    with pytest.raises(LJMCommunicationError):
        com.open()
    assert not com.is_open

    # opening T7-Pro version works same as opening T7
    working_config = dict(testconfig)
    for device_type in ("T7_PRO", LJMCommunicationConfig.DeviceType.T7_PRO):
        working_config["device_type"] = device_type
        assert not com.is_open
        with LJMCommunication(working_config) as com:
            assert com.is_open
        assert not com.is_open


def test_read_name(com) -> None:
    assert com.read_name("SERIAL_NUMBER") == 0.0
    assert com.read_name("SERIAL_NUMBER", "SERIAL_NUMBER") == [0.0, 0.0]
    sn_int = com.read_name("SERIAL_NUMBER", return_num_type=int)
    assert isinstance(sn_int, int)
    assert sn_int == 0

    with pytest.raises(TypeError):
        com.read_name(123)


def test_write_name(com) -> None:
    # Attention:
    # > With demo mode, functions like LJM_eReadNames return meaningless values. Future
    # > versions of LJM might return different values.
    com.write_name("test", 10)
    assert com.read_name("test") == 0.0

    with pytest.raises(LJMCommunicationError):
        com.write_name("blah", 0)

    com.write_names({"DAC0": 1, "AIN0": 2})
    read_values = com.read_name("DAC0", "AIN0")
    assert isinstance(read_values, abc.Sequence)
    assert len(read_values) == 2

    with pytest.raises(LJMCommunicationError):
        com.write_names({"blah": 0})

    with pytest.raises(TypeError):
        com.write_name(123, 456)

    with pytest.raises(TypeError):
        com.write_names({123: 456})


# def test_write_address(com):
#     with pytest.raises(LJMCommunicationError):
#         com.write_address(1234, 5678)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for the hvl_ccb.dev.ka3000p sub-package.
"""

import logging

import pytest

from hvl_ccb.dev.ka3000p import KA3000P, KA3000PError
from hvl_ccb.dev.ka3000p.base import KA3000PStatus, KA3000PTracking
from hvl_ccb.dev.ka3000p.comm import KA3000PCommunicationConfig
from masked_comm.serial import KA3000PLoopCommunication

logging.basicConfig(level=logging.ERROR)


@pytest.fixture(scope="module")
def com_config():
    return {
        "port": "loop://?logging=debug",
        "baudrate": 9600,
        "parity": KA3000PCommunicationConfig.Parity.NONE,
        "stopbits": KA3000PCommunicationConfig.Stopbits.ONE,
        "bytesize": KA3000PCommunicationConfig.Bytesize.EIGHTBITS,
        "terminator": b"\r\n",
        "timeout": 3,
        "wait_sec_read_text_nonempty": 0.01,
        "default_n_attempts_read_text_nonempty": 5,
    }


@pytest.fixture(scope="module")
def dev_config():
    return {}


@pytest.fixture
def started_ka3000p_device(com_config, dev_config):
    serial_port = KA3000PLoopCommunication(com_config)
    serial_port.open()
    serial_port.put_text("KORAD KA3005P V5.9 SN:02738566")
    with KA3000P(serial_port, dev_config) as ka3000p:
        while serial_port.get_written() is not None:
            pass
        yield serial_port, ka3000p


def gen_responce_from_status(
    status: KA3000PStatus | None = None,
    *,
    ch1_cv=False,
    ch2_cv=False,
    beep=False,
    ocp=False,
    output=False,
    ovp=False,
) -> bytes:
    status = status or KA3000PStatus(
        ch1_cv=ch1_cv,
        ch2_cv=ch2_cv,
        tracking=KA3000PTracking.SINGLE,
        beep=beep,
        ocp=ocp,
        output=output,
        ovp=ovp,
    )

    status_bits: int = 0
    status_bits += status.ch1_cv << 0
    status_bits += status.ch2_cv << 1
    status_bits += status.tracking.value << 2
    status_bits += status.beep << 4
    status_bits += status.ocp << 5
    status_bits += status.output << 6
    status_bits += status.ovp << 7

    return status_bits.to_bytes(length=1, byteorder="big")


def test_com_config(com_config) -> None:
    config = KA3000PCommunicationConfig(**com_config)
    for key, value in com_config.items():
        assert getattr(config, key) == value


def test_default_com(com_config) -> None:
    ka3000p = KA3000P(com_config)
    assert ka3000p


def test_broken_start(com_config, dev_config):
    serial_port = KA3000PLoopCommunication(com_config)

    ka3000p = KA3000P(serial_port, dev_config)
    assert not ka3000p.com.is_open
    serial_port.open()
    assert ka3000p.com.is_open
    with pytest.raises(KA3000PError):
        ka3000p.start()
    assert not ka3000p.com.is_open


def test_dev_init(
    started_ka3000p_device: tuple[KA3000PLoopCommunication, KA3000P],
) -> None:
    _com, ka3000p = started_ka3000p_device

    assert ka3000p.max_voltage == 30
    assert ka3000p.max_current == 5

    assert ka3000p.brand == "KORAD"
    assert ka3000p.comm_version == "V5.9"
    assert ka3000p.serial_number == "02738566"


def test_wrong_identifier(
    started_ka3000p_device: tuple[KA3000PLoopCommunication, KA3000P],
) -> None:
    com, ka3000p = started_ka3000p_device

    com.put_text("KORADKA3005PV5.9SN:02738566")
    with pytest.raises(KA3000PError):
        ka3000p.idenify_device()


def test_broken_com(
    started_ka3000p_device: tuple[KA3000PLoopCommunication, KA3000P],
) -> None:
    _com, ka3000p = started_ka3000p_device

    with pytest.raises(KA3000PError):
        ka3000p.idenify_device()


def test_voltage(
    started_ka3000p_device: tuple[KA3000PLoopCommunication, KA3000P],
) -> None:
    com, ka3000p = started_ka3000p_device

    # Read output voltage
    com.put_text("12.34")
    assert ka3000p.voltage == 12.34
    assert com.get_written() == "VOUT1?"

    # Set voltage
    ka3000p.voltage = 5.74
    assert com.get_written() == "VSET1:5.74"

    ka3000p.set_voltage = 4.56
    assert com.get_written() == "VSET1:4.56"

    # Read set voltage
    com.put_text("3.13")
    assert ka3000p.set_voltage == 3.13
    assert com.get_written() == "VSET1?"


def test_current(
    started_ka3000p_device: tuple[KA3000PLoopCommunication, KA3000P],
) -> None:
    com, ka3000p = started_ka3000p_device

    # Read output current
    com.put_text("1.234")
    assert ka3000p.current == 1.234
    assert com.get_written() == "IOUT1?"

    # Set current
    ka3000p.current = 4.74
    assert com.get_written() == "ISET1:4.740"

    ka3000p.set_current = 3.654
    assert com.get_written() == "ISET1:3.654"

    # Read set current
    com.put_text("0.135")
    assert ka3000p.set_current == 0.135
    assert com.get_written() == "ISET1?"


def test_status(
    started_ka3000p_device: tuple[KA3000PLoopCommunication, KA3000P],
) -> None:
    com, ka3000p = started_ka3000p_device

    status = KA3000PStatus(
        ch1_cv=False,
        ch2_cv=True,
        tracking=KA3000PTracking.SINGLE,
        beep=True,
        ocp=False,
        output=False,
        ovp=False,
    )
    status_byte = gen_responce_from_status(status)
    com.put_bytes(status_byte)
    assert ka3000p.status == status

    com.put_bytes(b"abc")
    with pytest.raises(KA3000PError):
        assert ka3000p.status

    com.put_bytes(b"")
    with pytest.raises(KA3000PError):
        assert ka3000p.status


def test_controlled_current_voltage(
    started_ka3000p_device: tuple[KA3000PLoopCommunication, KA3000P],
) -> None:
    com, ka3000p = started_ka3000p_device

    com.put_bytes(gen_responce_from_status(ch1_cv=False))
    assert ka3000p.controlled_current
    com.put_bytes(gen_responce_from_status(ch1_cv=False))
    assert not ka3000p.controlled_voltage

    com.put_bytes(gen_responce_from_status(ch1_cv=True))
    assert ka3000p.controlled_voltage
    com.put_bytes(gen_responce_from_status(ch1_cv=True))
    assert not ka3000p.controlled_current


def test_output(
    started_ka3000p_device: tuple[KA3000PLoopCommunication, KA3000P],
) -> None:
    com, ka3000p = started_ka3000p_device

    status_byte = gen_responce_from_status(status=None, output=False)
    com.put_bytes(status_byte)
    assert ka3000p.output is False
    assert com.get_written() == "STATUS?"

    status_byte = gen_responce_from_status(status=None, output=True)
    com.put_bytes(status_byte)
    assert ka3000p.output
    assert com.get_written() == "STATUS?"

    ka3000p.output = True
    assert com.get_written() == "OUT1"
    ka3000p.output = False
    assert com.get_written() == "OUT0"


def test_beep(
    started_ka3000p_device: tuple[KA3000PLoopCommunication, KA3000P],
) -> None:
    com, ka3000p = started_ka3000p_device

    status_byte = gen_responce_from_status(status=None, beep=False)
    com.put_bytes(status_byte)
    assert ka3000p.beep is False
    assert com.get_written() == "STATUS?"

    status_byte = gen_responce_from_status(status=None, beep=True)
    com.put_bytes(status_byte)
    assert ka3000p.beep
    assert com.get_written() == "STATUS?"

    ka3000p.beep = True
    assert com.get_written() == "BEEP1"
    ka3000p.beep = False
    assert com.get_written() == "BEEP0"


def test_lock(
    started_ka3000p_device: tuple[KA3000PLoopCommunication, KA3000P],
) -> None:
    com, ka3000p = started_ka3000p_device

    ka3000p.lock = True
    assert com.get_written() == "LOCK1"
    ka3000p.lock = False
    assert com.get_written() == "LOCK0"

    with pytest.raises(NotImplementedError):
        assert ka3000p.lock


def test_recall(
    started_ka3000p_device: tuple[KA3000PLoopCommunication, KA3000P],
) -> None:
    com, ka3000p = started_ka3000p_device

    ka3000p.recall = 3
    assert com.get_written() == "RCL3"

    with pytest.raises(ValueError):
        ka3000p.recall = -1
    with pytest.raises(ValueError):
        ka3000p.recall = 6
    with pytest.raises(TypeError):
        ka3000p.recall = 2.5

    with pytest.raises(NotImplementedError):
        assert ka3000p.recall


def test_save(
    started_ka3000p_device: tuple[KA3000PLoopCommunication, KA3000P],
) -> None:
    com, ka3000p = started_ka3000p_device
    ka3000p.save = 4
    assert com.get_written() == "SAV4"

    with ka3000p.save_settings_to(3):
        assert com.get_written() == "RCL3"
    assert com.get_written() == "SAV3"

    with pytest.raises(ValueError):
        ka3000p.save = -1
    with pytest.raises(ValueError):
        ka3000p.save = 6
    with pytest.raises(TypeError):
        ka3000p.save = 2.5

    with pytest.raises(NotImplementedError):
        assert ka3000p.save


def test_over_voltage_protection(
    started_ka3000p_device: tuple[KA3000PLoopCommunication, KA3000P],
) -> None:
    com, ka3000p = started_ka3000p_device

    status_byte = gen_responce_from_status(status=None, ovp=False)
    com.put_bytes(status_byte)
    assert not ka3000p.ovp
    assert com.get_written() == "STATUS?"

    status_byte = gen_responce_from_status(status=None, ovp=True)
    com.put_bytes(status_byte)
    assert ka3000p.ovp
    assert com.get_written() == "STATUS?"

    ka3000p.ovp = True
    assert com.get_written() == "OVP1"
    ka3000p.ovp = False
    assert com.get_written() == "OVP0"


def test_over_current_protection(
    started_ka3000p_device: tuple[KA3000PLoopCommunication, KA3000P],
) -> None:
    com, ka3000p = started_ka3000p_device

    status_byte = gen_responce_from_status(status=None, ocp=False)
    com.put_bytes(status_byte)
    assert not ka3000p.ocp
    assert com.get_written() == "STATUS?"

    status_byte = gen_responce_from_status(status=None, ocp=True)
    com.put_bytes(status_byte)
    assert ka3000p.ocp
    assert com.get_written() == "STATUS?"

    ka3000p.ocp = True
    assert com.get_written() == "OCP1"
    ka3000p.ocp = False
    assert com.get_written() == "OCP0"



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Testing Lauda PRO RP245E driver. Makes use of the socket library (TCP comm).
"""

from collections.abc import Generator

import pytest

from hvl_ccb.dev.lauda import (
    LaudaProRp245e,
    LaudaProRp245eCommand,
    LaudaProRp245eCommandError,
    LaudaProRp245eConfig,
    LaudaProRp245eTcpCommunicationConfig,
)
from mocked_comm.tcp import LocalLaudaServer, get_free_tcp_port


@pytest.fixture
def com_config() -> dict[str, str | float | int | bytes]:
    host = "127.0.0.1"
    return {
        "host": host,
        "port": get_free_tcp_port(host),
        "wait_sec_read_text_nonempty": 0.01,
        "default_n_attempts_read_text_nonempty": 2,
        "terminator": b"\r\n",
        "timeout": 0.1,
    }


@pytest.fixture(scope="module")
def dev_config():
    return {
        "operation_mode": 2,
        "control_mode": 0,
        "upper_temp": 200,
        "lower_temp": -50,
    }


@pytest.fixture
def started_lauda_server(
    com_config, dev_config
) -> Generator[tuple[LocalLaudaServer, LaudaProRp245e], None, None]:
    ts = LocalLaudaServer(
        port=com_config["port"],
        terminator=com_config["terminator"],
        timeout=com_config["timeout"],
    )
    lauda = LaudaProRp245e(com_config, dev_config)
    with ts, lauda:
        ts._starting.join()
        yield ts, lauda


def test_com_config(com_config) -> None:
    config = LaudaProRp245eTcpCommunicationConfig(**com_config)
    for key, value in com_config.items():
        assert getattr(config, key) == value


@pytest.mark.parametrize(
    "wrong_config_dict",
    [
        {"wait_sec_pre_read_or_write": -0.5},
        {"terminator": b"NotATerminator"},
    ],
)
def test_invalid_com_config_dict(com_config, wrong_config_dict) -> None:
    invalid_config = dict(com_config)
    invalid_config.update(wrong_config_dict)
    with pytest.raises(ValueError):
        LaudaProRp245eTcpCommunicationConfig(**invalid_config)


def test_dev_config(dev_config) -> None:
    # currently there are no non-default config values

    lauda_config = LaudaProRp245eConfig()
    assert lauda_config

    config = LaudaProRp245eConfig(**dev_config)
    for key, value in dev_config.items():
        assert getattr(config, key) == value


@pytest.mark.parametrize(
    "wrong_config_dict",
    [
        {"operation_mode": 123},
        {"control_mode": 123},
        {"lower_temp": -100},
        {"upper_temp": 300},
    ],
)
def test_invalid_config_dict(dev_config, wrong_config_dict) -> None:
    invalid_config = dict(dev_config)
    invalid_config.update(wrong_config_dict)
    with pytest.raises(ValueError):
        LaudaProRp245eConfig(**invalid_config)


def test_started_devices(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    assert lauda.get_device_type() == "RP245 PRO"


def test_get_device_type(started_lauda_server) -> None:
    ts, lauda = started_lauda_server

    assert lauda.get_device_type() == "RP245 PRO"

    ts.custom_answer = "RP245"  # without the magic "PRO"...
    with pytest.raises(LaudaProRp245eCommandError):
        lauda.get_device_type()


def test_pause(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    assert lauda.pause() == LaudaProRp245eCommand.STOP


def test_read_error(started_lauda_server) -> None:
    ts, lauda = started_lauda_server

    ts.custom_answer = "ERR"
    with pytest.raises(LaudaProRp245eCommandError):
        lauda.start_ramp()


def test_run(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    assert lauda.run() == LaudaProRp245eCommand.START


def test_set_external_temp(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    ext = 20.00
    assert (
        lauda.set_external_temp(ext)
        == f"{LaudaProRp245eCommand.EXTERNAL_TEMP}{ext:.2f}"
    )


def test_set_temp_set_point(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    ext = 25.00
    assert (
        lauda.set_temp_set_point(ext)
        == f"{LaudaProRp245eCommand.TEMP_SET_POINT}{ext:.2f}"
    )


def test_correct_pump_level(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    ext = 3
    assert lauda.set_pump_level(ext) == f"{LaudaProRp245eCommand.PUMP_LEVEL}{ext}"


def test_wrong_pump_level(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    ext = 15
    with pytest.raises(ValueError):
        lauda.set_pump_level(ext)


def test_correct_control_mode(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    ext = 0
    assert lauda.set_control_mode(0) == f"{LaudaProRp245eCommand.CONT_MODE}{ext}"


def test_wrong_control_mode(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    with pytest.raises(ValueError):
        lauda.set_control_mode(6)


def test_correct_ramp_program(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    ext = 1
    assert lauda.set_ramp_program(ext) == f"{LaudaProRp245eCommand.RAMP_SELECT}{ext}"


def test_wrong_ramp_program(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    with pytest.raises(ValueError):
        lauda.set_ramp_program(6)


def test_start_ramp(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    assert lauda.start_ramp() == LaudaProRp245eCommand.RAMP_START


def test_pause_ramp(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    assert lauda.pause_ramp() == LaudaProRp245eCommand.RAMP_PAUSE


def test_continue_ramp(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    assert lauda.continue_ramp() == LaudaProRp245eCommand.RAMP_CONTINUE


def test_stop_ramp(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    assert lauda.stop_ramp() == LaudaProRp245eCommand.RAMP_STOP


def test_set_ramp_iterations(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    ext = 3
    assert (
        lauda.set_ramp_iterations(ext)
        == f"{LaudaProRp245eCommand.RAMP_ITERATIONS}{ext}"
    )


def test_reset_ramp(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    assert lauda.reset_ramp() == LaudaProRp245eCommand.RAMP_DELETE


def test_set_ramp_segment(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    temp = 10.00
    dur = 0
    tol = 0.00
    pump = 3
    segment = f"{temp:.2f}_{dur}_{tol:.2f}_{pump}"
    assert (
        lauda.set_ramp_segment(temp, dur, tol, pump)
        == f"{LaudaProRp245eCommand.RAMP_SET}{segment}"
    )


def test_get_bath_temp(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    assert lauda.get_bath_temp() == 25.00


def test_get_ext_temp(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    assert lauda.get_external_temp() == 25.00


def test_returns_error(started_lauda_server) -> None:
    _ts, lauda = started_lauda_server

    with pytest.raises(LaudaProRp245eCommandError):
        lauda.com.query_command("FOOBAR")



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Device classes for "Probus V - ADDAT30" Interfaces which are used to control power
supplies from FuG Elektronik GmbH

This interface is used for many FuG power units.
Manufacturer homepage:
https://www.fug-elektronik.de

The Professional Series of Power Supplies from FuG is a series of low, medium and high
voltage direct current power supplies as well as capacitor chargers.
The class FuG is tested with a HCK 800-20 000 in Standard Mode.
The addressable mode is not implemented.
Check the code carefully before using it with other devices.
Manufacturer homepage:
https://www.fug-elektronik.de/netzgeraete/professional-series/

The documentation of the interface from the manufacturer can be found here:
https://www.fug-elektronik.de/wp-content/uploads/download/de/SOFTWARE/Probus_V.zip

The provided classes support the basic and some advanced commands.
The commands for calibrating the power supplies are not implemented, as they are only
for very special porpoises and should not used by "normal" customers.
"""

import logging
import re
from typing import cast

from hvl_ccb.configuration import configdataclass
from hvl_ccb.dev.base import SingleCommDevice
from hvl_ccb.utils.typing import Number

from .comm import FuGSerialCommunication
from .constants import FuGDigitalVal, FuGProbusIVCommands
from .errors import FuGError, FuGErrorcodes
from .registers import (
    FuGProbusVConfigRegisters,
    FuGProbusVDIRegisters,
    FuGProbusVDORegisters,
    FuGProbusVMonitorRegisters,
    FuGProbusVRegisterGroups,
    FuGProbusVSetRegisters,
)

logger = logging.getLogger(__name__)


@configdataclass
class FuGConfig:
    """
    Device configuration dataclass for FuG power supplies.
    """


class FuGProbusIV(SingleCommDevice):
    """
    FuG Probus IV device class

    Sends basic SCPI commands and reads the answer.
    Only the special commands and PROBUS IV instruction set is implemented.
    """

    def __init__(self, com, dev_config=None) -> None:
        # Call superclass constructor
        super().__init__(com, dev_config)

        # Version of the interface (will be retrieved after com is opened)
        self._interface_version: str = ""

    def __repr__(self) -> str:
        return f"FuGProbus({self._interface_version})"

    @staticmethod
    def default_com_cls():
        return FuGSerialCommunication

    @staticmethod
    def config_cls():
        return FuGConfig

    def start(self) -> None:
        logger.info(f"Starting device {self}")
        super().start()

        self._interface_version = self.command(FuGProbusIVCommands.ID)  # type: ignore[arg-type]
        logger.info(f"Connection to {self._interface_version} established.")

    def stop(self) -> None:
        with self.com.access_lock:
            logger.info(f"Stopping device {self}")
            self.output_off()
            self.reset()
            super().stop()

    def command(self, command: FuGProbusIVCommands, value=None) -> str:
        """

        :param command: one of the commands given within FuGProbusIVCommands
        :param value: an optional value, depending on the command
        :return: a String if a query was performed
        """
        if not (
            (value is None and command.input_type is None)
            or isinstance(value, command.input_type)
        ):
            msg = (
                "Wrong value for data was given. Expected: "
                f"{command.input_type} and given: {value.__class__}"
            )
            raise FuGError(msg)

        # Differentiate between with and without optional value
        if command.input_type is None:
            answer = self.com.query(f"{command.command}")
        else:
            answer = self.com.query(f"{command.command}{value}")
        return cast("str", answer)  # for typing: query of fug never returns None

    # Special commands
    def reset(self) -> None:
        """
        Reset of the interface:
        All setvalues are set to zero
        """
        self.command(FuGProbusIVCommands.RESET)  # type: ignore[arg-type]

    def output_off(self) -> None:
        """
        Switch DC voltage output off.
        """
        self.command(
            FuGProbusIVCommands.OUTPUT,  # type: ignore[arg-type]
            FuGDigitalVal.OFF,
        )


class FuGProbusV(FuGProbusIV):
    """
    FuG Probus V class which uses register based commands to control the power supplies
    """

    def set_register(self, register: str, value: Number | str) -> None:
        """
        generic method to set value to register

        :param register: the name of the register to set the value
        :param value: which should be written to the register
        """

        self.com.query(f">{register} {value}")

    def get_register(self, register: str) -> str:
        """
        get the value from a register

        :param register: the register from which the value is requested
        :returns: the value of the register as a String
        """

        answer = cast("str", self.com.query(f">{register} ?")).split(":")
        # cast for typing: query of fug never returns None

        if answer[0] != register:
            cast("FuGErrorcodes", FuGErrorcodes.E505).raise_()

        return answer[1]


class FuG(FuGProbusV):
    """
    FuG power supply device class.

    The power supply is controlled over a FuG ADDA Interface with the PROBUS V protocol
    """

    def __init__(self, com, dev_config=None) -> None:
        """
        Constructor for configuring the power supply.

        :param com:
        :param dev_config:
        """

        # Call superclass constructor
        super().__init__(com, dev_config)

        self._id_string: str = ""
        """ID String of the device (will be retrieved after com is opened) contains
        Serial number and model"""

        # Serial number of the device (will be retrieved after com is opened)
        self._serial_number: str = ""
        # model class of the device (derived from serial number)
        self._model: str = ""
        # maximum output current of the hardware
        self._max_current_hardware: Number = 0
        # maximum output charging power of the hardware
        self._max_power_hardware: Number = 0
        # maximum output voltage of the hardware
        self._max_voltage_hardware: Number = 0

        self._voltage = FuGProbusVSetRegisters(
            self, FuGProbusVRegisterGroups("SETVOLTAGE")
        )
        self._current = FuGProbusVSetRegisters(
            self, FuGProbusVRegisterGroups("SETCURRENT")
        )
        self._out_x0 = FuGProbusVDORegisters(self, FuGProbusVRegisterGroups("OUTPUTX0"))
        self._out_x1 = FuGProbusVDORegisters(self, FuGProbusVRegisterGroups("OUTPUTX1"))
        self._out_x2 = FuGProbusVDORegisters(self, FuGProbusVRegisterGroups("OUTPUTX2"))
        self._out_xcmd = FuGProbusVDORegisters(
            self, FuGProbusVRegisterGroups("OUTPUTXCMD")
        )
        self._output = FuGProbusVDORegisters(
            self, FuGProbusVRegisterGroups("OUTPUTONCMD")
        )
        self._voltage_monitor = FuGProbusVMonitorRegisters(
            self, FuGProbusVRegisterGroups("MONITOR_V")
        )
        self._current_monitor = FuGProbusVMonitorRegisters(
            self, FuGProbusVRegisterGroups("MONITOR_I")
        )
        self._di = FuGProbusVDIRegisters(self, FuGProbusVRegisterGroups("INPUT"))
        self._config_status = FuGProbusVConfigRegisters(
            self, FuGProbusVRegisterGroups("CONFIG")
        )

    def __repr__(self) -> str:
        return f"{self._id_string}"

    def start(self, max_voltage=0, max_current=0) -> None:
        """
        Opens the communication protocol and configures the device.

        :param max_voltage: Configure here the maximal permissible voltage which is
            allowed in the given experimental setup
        :param max_current: Configure here the maximal permissible current which is
            allowed in the given experimental setup
        """

        # starting FuG Probus Interface
        super().start()

        self._voltage._max_setvalue = max_voltage
        self._current._max_setvalue = max_current

        # find out which type of source this is:
        self.identify_device()

    @property
    def voltage(self) -> Number:
        """
        Return the measured output voltage in V
        """
        return self.voltage_monitor.value

    @voltage.setter
    def voltage(self, value: Number) -> None:
        """
        Set the output voltage in V

        :param value: voltage in V
        """
        self._voltage.setvalue = value

    @property
    def current(self) -> Number:
        """
        Return the measured output current in A
        """
        return self.current_monitor.value

    @current.setter
    def current(self, value: Number) -> None:
        """
        Set the output current

        :param value: Current in A
        """
        self._current.setvalue = value

    @property
    def set_voltage(self) -> Number:
        """
        Return the set voltage (may differ from actual voltage) in V
        """
        return self._voltage.setvalue

    @set_voltage.setter
    def set_voltage(self, value: Number) -> None:
        """Set the output voltage"""
        self.voltage = value

    @property
    def set_current(self) -> Number:
        """Return the set current (may differ from actual value) in A"""
        return self._current.setvalue

    @set_current.setter
    def set_current(self, value: Number) -> None:
        """Set the output current"""
        self.current = value

    @property
    def max_voltage(self) -> Number:
        """
        Returns the maximal voltage which could provided within the test setup

        :return: max voltage in V
        """
        return self._voltage._max_setvalue

    @property
    def max_voltage_hardware(self) -> Number:
        """
        Returns the maximal voltage which could provided with the power supply

        :return:
        """
        return self._max_voltage_hardware

    @property
    def max_current(self) -> Number:
        """
        Returns the maximal current which could provided within the test setup

        :return: max current in A
        """
        return self._current._max_setvalue

    @property
    def max_current_hardware(self) -> Number:
        """
        Returns the maximal current which could provided with the power supply

        :return:
        """
        return self._max_current_hardware

    # Output stage
    @property
    def output(self) -> bool | None:
        """State of the high voltage output"""
        return self._output.status == FuGDigitalVal.ON

    @output.setter
    def output(self, value: bool) -> None:
        """
        Activates the output of the source

        :param value: `True` for activation, `False` for deactivation
        """
        self._output.out = FuGDigitalVal.ON if value else FuGDigitalVal.OFF

    # Utility Registers (Not implementing `Source`-protocol)
    @property
    def voltage_register(self) -> FuGProbusVSetRegisters:
        """
        Returns the registers for the voltage output
        """
        return self._voltage

    @property
    def current_register(self) -> FuGProbusVSetRegisters:
        """
        Returns the registers for the current output
        """
        return self._current

    @property
    def out_x0(self) -> FuGProbusVDORegisters:
        """
        Returns the registers for the digital output X0

        :return: FuGProbusVDORegisters
        """
        return self._out_x0

    @property
    def out_x1(self) -> FuGProbusVDORegisters:
        """
        Returns the registers for the digital output X1

        :return: FuGProbusVDORegisters
        """
        return self._out_x1

    @property
    def out_x2(self) -> FuGProbusVDORegisters:
        """
        Returns the registers for the digital output X2

        :return: FuGProbusVDORegisters
        """
        return self._out_x2

    @property
    def out_xcmd(self) -> FuGProbusVDORegisters:
        """
        Returns the registers for the digital outputX-CMD

        :return: FuGProbusVDORegisters
        """
        return self._out_xcmd

    @property
    def output_register(self) -> FuGProbusVDORegisters:
        """
        Returns the registers for the output switch to turn the output on or off

        :return: FuGProbusVDORegisters
        """
        return self._output

    @property
    def voltage_monitor(self) -> FuGProbusVMonitorRegisters:
        """
        Returns the registers for the voltage monitor.

        A typically usage will be "self.voltage_monitor.value" to measure the output
        voltage

        :return:
        """
        return self._voltage_monitor

    @property
    def current_monitor(self) -> FuGProbusVMonitorRegisters:
        """
        Returns the registers for the current monitor.

        A typically usage will be "self.current_monitor.value" to measure the output
        current

        :return:
        """
        return self._current_monitor

    @property
    def di(self) -> FuGProbusVDIRegisters:
        """
        Returns the registers for the digital inputs

        :return: FuGProbusVDIRegisters
        """
        return self._di

    @property
    def config_status(self) -> FuGProbusVConfigRegisters:
        """
        Returns the registers for the registers with the configuration and status values

        :return: FuGProbusVConfigRegisters
        """
        return self._config_status

    def identify_device(self) -> None:
        """
        Identify the device nominal voltage and current based on its model number.

        :raises SerialCommunicationIOError: when communication port is not opened
        """
        id_string = str(self.command(FuGProbusIVCommands("ID")))
        # "'FUG HCK
        # 800
        # - 20 000
        # MOD 17022-01-01'"
        # regex to find the model of the device
        regex_model = (
            "FUG (?P<model>[A-Z]{3})"
            " (?P<power>[0-9 ]+)"
            " - (?P<voltage>[0-9 ]+)"
            " MOD (?P<sn>[0-9-]+)"
        )

        result = re.search(regex_model, id_string)
        if not result:
            msg = (
                f'The device with the ID string "{id_string}" could not be recognized.'
            )
            raise FuGError(msg)

        self._id_string = id_string
        results = result.groupdict()
        self._model = str(results.get("model"))
        self._max_power_hardware = int(results.get("power").replace(" ", ""))  # type: ignore[union-attr]
        self._max_voltage_hardware = int(results.get("voltage").replace(" ", ""))  # type: ignore[union-attr]
        self._max_current_hardware = (
            2 * self._max_power_hardware / self._max_voltage_hardware
        )
        self._serial_number = str(results.get("sn"))

        logger.info(f"Device {id_string} successfully identified:")
        logger.info(f"Model class: {self._model}")
        logger.info(f"Maximal voltage: {self._max_voltage_hardware} V")
        logger.info(f"Maximal current: {self._max_current_hardware} A")
        logger.info(f"Maximal charging power: {self._max_power_hardware} J/s")
        logger.info(f"Serial number: {self._serial_number}")

        # if limits for test setup were not predefined, set them to hardware limits
        # or if the previous limits were to high, limit them to the hardware limits
        if self.max_voltage == 0:
            self._voltage._max_setvalue = self.max_voltage_hardware
        elif self.max_voltage > self.max_voltage_hardware:
            logger.warning(
                "FuG power source should supply up to "
                f"{self.max_voltage} V, but the hardware only goes up "
                f"to {self.max_voltage_hardware} V."
            )
            self._voltage._max_setvalue = self.max_voltage_hardware
        logger.info(
            "For this setup the maximal output voltage of the power "
            f"supply is limited to {self.max_voltage} V."
        )

        if self.max_current == 0:
            self._current._max_setvalue = self.max_current_hardware
        elif self.max_current > self.max_current_hardware:
            logger.warning(
                "FuG power source should supply up to "
                f"{self.max_current} A, but the hardware only goes up "
                f"to {self.max_current_hardware} A."
            )
            self._current._max_setvalue = self.max_current_hardware
        logger.info(
            "For this setup the maximal output current of the power "
            f"supply is limited to {self.max_current} A."
        )



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
This file is only for type checking, that the `Source`-protocol is fulfilled.
"""

from hvl_ccb.dev.fug import FuG
from hvl_ccb.dev.heinzinger import Heinzinger
from hvl_ccb.dev.ka3000p import KA3000P
from hvl_ccb.dev.technix import Technix

from . import Source

my_source: Source = Heinzinger({})
my_source = Technix({}, {})
my_source = FuG({})
my_source = KA3000P({})



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
from typing import Protocol, runtime_checkable

from hvl_ccb.utils.typing import Number


@runtime_checkable
class Source(Protocol):
    """
    Protocol for a voltage/current source

    (Future) Implementations:
        - `ea_psi9000.PSI9000`
        - `fug.Fug`
        - `heinzinger.HeinzingerPNC`
        - `technix.Technix`
        - `ka3000p.KA3000P`
    """

    # maximum output voltage of the hardware
    _max_voltage_hardware: Number
    # maximum output current of the hardware
    _max_current_hardware: Number

    # Voltage and current values
    @property
    def voltage(self) -> Number:
        """Return the measured output voltage in V"""
        return 0

    @voltage.setter
    def voltage(self, value: Number) -> None:
        """Set the output voltage"""

    @property
    def current(self) -> Number:
        """Return the measured output current in A"""
        return 0

    @current.setter
    def current(self, value: Number) -> None:
        """Set the output current"""

    @property
    def set_voltage(self) -> Number:
        """Return the set voltage (may differ from actual value) in V"""
        return 0

    @set_voltage.setter
    def set_voltage(self, value: Number) -> None:
        """Set the output voltage"""
        self.voltage = value

    @property
    def set_current(self) -> Number:
        """Return the set current (may differ from actual value) in A"""
        return 0

    @set_current.setter
    def set_current(self, value: Number) -> None:
        """Set the output current"""
        self.current = value

    @property
    def max_voltage(self) -> Number:
        """Maximal output voltage of the hardware in V,
        but user can reset to a lower value"""
        return self._max_voltage_hardware

    @property
    def max_current(self) -> Number:
        """Maximal output current of the hardware in A,
        but user can reset to a lower value"""
        return self._max_current_hardware

    # Output stage
    @property
    def output(self) -> bool | None:
        """State of the high voltage output"""
        return False

    @output.setter
    def output(self, value: bool) -> None:
        """
        Activates the output of the source

        :param value: `True` for activation, `False` for deactivation
        """



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
""" """

import logging
import socket
from collections.abc import Sequence
from ipaddress import IPv4Address, IPv6Address, ip_address
from logging import Logger

import numpy as np

from hvl_ccb.utils.typing import Number


def validate_number(
    x_name: str,
    x: object,
    limits: tuple | None = (None, None),
    number_type: type[Number] | tuple[type[Number], ...] = (int, float),
    logger: Logger | None = None,
) -> None:
    """
    Validate if given input `x` is a number of given `number_type` type, with value
    between given `limits[0]` and `limits[1]` (inclusive), if not `None`.
    For array-like objects (npt.NDArray, list, tuple, dict) it is checked if all
    elements are within the limits and have the correct type.

    :param x_name: string name of the validate input, use for the error message
    :param x: an input object to validate as number of given type within given range
    :param logger: logger of the calling submodule
    :param limits: [lower, upper] limit, with `None` denoting no limit: [-inf, +inf]
    :param number_type: expected type or tuple of types of a number,
        by default `(int, float)`
    :raises TypeError: when the validated input does not have expected type
    :raises ValueError: when the validated input has correct number type but is not
        within given range or has wrong input limits
    """
    if logger is None:
        logger = logging.getLogger(__name__)

    if limits is None:
        limits = (-np.inf, np.inf)
    if limits[0] is None:
        limits = (-np.inf, limits[1])
    if limits[1] is None:
        limits = (limits[0], np.inf)
    if limits[0] >= limits[1]:
        msg = (
            f"Upper limit {limits[1]} should be greater than "
            f"the lower limit {limits[0]}"
        )
        logger.error(msg)
        raise ValueError(msg)

    data_type = type(x)
    if not isinstance(number_type, Sequence):
        number_type = (number_type,)

    if isinstance(x, float | int):
        if not isinstance(x, number_type):
            msg = (
                f"{x_name} = {x} has to be of type "
                f"{' or '.join(nt.__name__ for nt in number_type)}"
            )
            logger.error(msg)
            raise TypeError(msg)
    elif isinstance(x, list | tuple | dict | np.ndarray):
        if isinstance(x, dict):
            x = np.asarray(list(x.values()))
        x = np.asarray(x)
        if x.dtype not in number_type:
            msg = (
                f"{x_name} = {x} needs to include only numbers type "
                f"{' or '.join(nt.__name__ for nt in number_type)}"
            )
            logger.error(msg)
            raise TypeError(msg)
    else:
        msg = (
            f"{x_name} = {x} must be an Integer, a Float, a Tuple, a List, "
            f"a Dictionary or a Numpy array, but the received type is {data_type}."
        )
        logger.error(msg)
        raise TypeError(msg)

    if np.any(x < limits[0]) or np.any(x > limits[1]):
        if np.isinf(limits[0]):
            suffix = f"less or equal than {limits[1]}"
        elif np.isinf(limits[1]):
            suffix = f"greater or equal than {limits[0]}"
        else:
            suffix = f"between {limits[0]} and {limits[1]} inclusive"
        msg = f"{x_name} = {x} has to be {suffix}"
        logger.error(msg)
        raise ValueError(msg)


def validate_bool(x_name: str, x: object, logger: Logger | None = None) -> None:
    """
    Validate if given input `x` is a `bool`.

    :param x_name: string name of the validate input, use for the error message
    :param x: an input object to validate as boolean
    :param logger: logger of the calling submodule
    :raises TypeError: when the validated input does not have boolean type
    """
    if logger is None:
        logger = logging.getLogger(__name__)
    if not isinstance(x, bool):
        msg = f"{x_name} = {x} has to of type bool"
        logger.error(msg)
        raise TypeError(msg)


def validate_and_resolve_host(
    host: str | IPv4Address | IPv6Address | None, logger: Logger | None = None
) -> str:
    if logger is None:
        logger = logging.getLogger(__name__)
    if host is None:
        msg = "A host has to be provided."
        logger.error(msg)
        raise AttributeError(msg)
    if isinstance(host, IPv4Address | IPv6Address):
        host = str(host)
    else:
        try:
            host = str(ip_address(host))
        except ValueError:
            try:
                host = socket.gethostbyname(host)
            except (socket.gaierror, UnicodeError) as exc:
                # UnicodeError: "192..168.01"
                # socket.gaierror: "192.168.0.1000"
                # socket.gaierror: "itet-hvl-01"
                msg = (
                    "The value of host is neither an IPv4 nor an IPv6 address nor a "
                    f"hostname that can be resolved. The received value is: '{host}'."
                )
                logger.exception(msg, exc_info=exc)
                raise ValueError(msg) from exc
    return str(host)


def validate_tcp_port(port: int | None, logger: Logger | None = None) -> None:
    if logger is None:
        logger = logging.getLogger(__name__)
    if port is None:
        msg = "A port has to be provided."
        logger.error(msg)
        raise AttributeError(msg)
    # TCP ports are in the range between and including 0 and (2**16)-1, but 0 is
    # reserved and shall not be used
    validate_number("port", port, (1, 2**16 - 1), int, logger)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Module with base classes for devices.
"""

import logging
from abc import ABC, abstractmethod

from typing_extensions import Self

from hvl_ccb.comm import CommunicationProtocol
from hvl_ccb.configuration import ConfigurationMixin, configdataclass
from hvl_ccb.error import CCBError

logger = logging.getLogger(__name__)


class DeviceError(CCBError):
    pass


class DeviceExistingError(DeviceError):
    """
    Error to indicate that a device with that name already exists.
    """


class DeviceFailuresError(DeviceError):
    """
    Error to indicate that one or several devices failed.
    """

    def __init__(self, failures: dict[str, Exception]) -> None:
        super().__init__(failures)
        self.failures: dict[str, Exception] = failures
        """A dictionary of named devices failures (exceptions).
        """


@configdataclass
class EmptyConfig:
    """
    Empty configuration dataclass that is the default configuration for a Device.
    """


class Device(ConfigurationMixin, ABC):
    """
    Base class for devices. Implement this class for a concrete device,
    such as measurement equipment or voltage sources.

    Specifies the methods to implement for a device.
    """

    def __init__(self, dev_config=None) -> None:
        """
        Constructor for Device.
        """

        super().__init__(dev_config)

    @abstractmethod
    def start(self) -> None:
        """
        Start or restart this Device. To be implemented in the subclass.
        """

    @abstractmethod
    def stop(self) -> None:
        """
        Stop this Device. To be implemented in the subclass.
        """

    def __enter__(self) -> Self:
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.stop()

    @staticmethod
    def config_cls():
        return EmptyConfig


class DeviceSequenceMixin:
    """
    Mixin that can be used on a device or other classes to provide facilities for
    handling multiple devices in a sequence.
    """

    def __init__(self, devices: dict[str, Device]) -> None:
        """
        Constructor for the DeviceSequenceMixin.

        :param devices: is a dictionary of devices to be added to this sequence.
        """

        super().__init__()

        self._devices: dict[str, Device] = {}
        for name, device in devices.items():
            self.add_device(name, device)

        self.devices_failed_start: dict[str, Device] = {}
        """Dictionary of named device instances from the sequence for which the most
        recent `start()` attempt failed.

        Empty if `stop()` was called last; cf. `devices_failed_stop`."""
        self.devices_failed_stop: dict[str, Device] = {}
        """Dictionary of named device instances from the sequence for which the most
        recent `stop()` attempt failed.

        Empty if `start()` was called last; cf. `devices_failed_start`."""

    def __getattribute__(self, item: str) -> Device | object:
        """
        Gets Device from the sequence or object's attribute.

        :param item: Item name to get
        :return: Device or object's attribute.
        """
        if item != "_devices" and item in self._devices:
            return self.get_device(item)
        return super().__getattribute__(item)

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, DeviceSequenceMixin) and self._devices == other._devices
        )

    def get_devices(self) -> list[tuple[str, Device]]:
        """
        Get list of name, device pairs according to current sequence.

        :return: A list of tuples with name and device each.
        """
        return list(self._devices.items())

    def get_device(self, name: str) -> Device:
        """
        Get a device by name.

        :param name: is the name of the device.
        :return: the device object from this sequence.
        """

        return self._devices[name]

    def add_device(self, name: str, device: Device) -> None:
        """
        Add a new device to the device sequence.

        :param name: is the name of the device.
        :param device: is the instantiated Device object.
        :raise DeviceExistingError:
        """

        if name in self._devices:
            raise DeviceExistingError

        # disallow over-shadowing via ".DEVICE_NAME" lookup
        if hasattr(self, name):
            msg = (
                "This sequence already has an attribute called"
                f" {name}. Use different name for this device."
            )
            raise ValueError(msg)

        self._devices[name] = device

    def remove_device(self, name: str) -> Device:
        """
        Remove a device from this sequence and return the device object.

        :param name: is the name of the device.
        :return: device object or `None` if such device was not in the sequence.
        :raises ValueError: when device with given name was not found
        """

        if name not in self._devices:
            msg = f'No device named "{name}" in this sequence.'
            raise ValueError(msg)

        if name in self.devices_failed_start:
            self.devices_failed_start.pop(name)
        elif name in self.devices_failed_stop:
            self.devices_failed_stop.pop(name)

        return self._devices.pop(name)

    def start(self) -> None:
        """
        Start all devices in this sequence in their added order.

        :raises DeviceFailuresError: if one or several devices failed to start
        """

        # reset the failure dicts
        failures = {}
        self.devices_failed_start = {}
        self.devices_failed_stop = {}

        for name, device in self._devices.items():
            try:
                device.start()
            except Exception as exc:  # noqa: PERF203 Continue loop after exception
                logger.exception(f"Could not start {name}: ", exc_info=exc)
                failures[name] = exc
                self.devices_failed_start[name] = device
        if failures:
            raise DeviceFailuresError(failures)

    def stop(self) -> None:
        """
        Stop all devices in this sequence in their reverse order.

        :raises DeviceFailuresError: if one or several devices failed to stop
        """

        # reset the failure dicts
        failures: dict[str, Exception] = {}
        self.devices_failed_start = {}
        self.devices_failed_stop = {}

        for name, device in self._devices.items():
            try:
                device.stop()
            except Exception as exc:  # noqa: PERF203 Continue loop after exception
                logger.exception(f"Could not start {name}: ", exc_info=exc)
                failures[name] = exc
                self.devices_failed_stop[name] = device
        if failures:
            raise DeviceFailuresError(failures)


class SingleCommDevice(Device, ABC):
    """
    Base class for devices with a single communication protocol.
    """

    # Omitting typing hint `com: CommunicationProtocol` on purpose
    # to enable PyCharm autocompletion for subtypes.
    def __init__(self, com, dev_config=None) -> None:
        """
        Constructor for Device. Links the communication protocol and provides a
        configuration for the device.

        :param com: Communication protocol to be used with
            this device. Can be of type: - CommunicationProtocol instance, - dictionary
            with keys and values to be used as configuration together with the
            default communication protocol, or - @configdataclass to be used together
            with the default communication protocol.

        :param dev_config: configuration of the device. Can be:
            - None: empty configuration is used, or the specified config_cls()
            - @configdataclass decorated class
            - Dictionary, which is then used to instantiate the specified config_cls()
        """

        super().__init__(dev_config)

        if isinstance(com, CommunicationProtocol):
            self._com = com
        else:
            self._com = self.default_com_cls()(com)

    @staticmethod
    @abstractmethod
    def default_com_cls() -> type[CommunicationProtocol]:
        """
        Get the class for the default communication protocol used with this device.

        :return: the type of the standard communication protocol for this device
        """

    @property
    def com(self):  # noqa: ANN201
        """
        Get the communication protocol of this device.

        :return: an instance of CommunicationProtocol subtype
        """
        return self._com

    def start(self) -> None:
        """
        Open the associated communication protocol.
        """

        self.com.open()

    def stop(self) -> None:
        """
        Close the associated communication protocol.
        """

        self.com.close()



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Constants, variable names for the BaseCube OPC-connected devices.
"""

from __future__ import annotations

import logging
from abc import abstractmethod
from typing import TYPE_CHECKING, cast

from aenum import Enum, IntEnum
from typing_extensions import Self

from hvl_ccb.utils.enum import ValueEnum, unique
from hvl_ccb.utils.validation import validate_bool

if TYPE_CHECKING:
    from collections.abc import Sequence, Sized

    from hvl_ccb.utils.typing import Number

    from . import BaseCube  # pragma: no cover

logger = logging.getLogger(__name__)


class _BaseGetDescriptor:
    """
    Base Descriptor for attributes that are readonly.
    """

    def __init__(self, number: Number, name: str) -> None:
        self.number = number
        self.name = name

    @abstractmethod
    def __get__(self, instance, owner):  # noqa: ANN204
        pass  # pragma: no cover

    def __set__(self, instance, state) -> None:
        msg = f"It is not possible to set the {self.name}. This parameter is readonly!"
        logger.error(msg)
        raise AttributeError(msg)


@unique
class _CubeOpcEndpoint(ValueEnum):
    """
    OPC Server Endpoint strings for the BaseCube variants.
    """

    PI_CUBE = "PICube"
    BASE_CUBE = "BaseCube"


@unique
class _BreakdownDetection(ValueEnum):
    """
    Node ID strings for the breakdown detection.
    """

    # Boolean read-only variable indicating whether breakdown detection and fast
    # switchoff is enabled in the system or not.
    ACTIVATED = '"DB_Fast_Switch_Off"."sx_breakdownD_active"'

    # Boolean read-only variable telling whether the fast switch-off has triggered.
    # This can also be seen using the safety circuit state, therefore no method is
    # implemented to read this out directly.
    TRIGGERED = '"DB_Fast_Switch_Off"."sx_breakdownD_triggered"'

    # Boolean writable variable to reset the fast switch-off. Toggle to re-enable.
    RESET = '"DB_Fast_Switch_Off"."sx_breakdownD_reset"'


_CEE16 = '"Qx_Allg_Socket_CEE16"'


class _T13Socket:
    """
    Set and get the state of a SEV T13 power socket.
    """

    _SOCKETS: tuple = (1, 2, 3)

    def __init__(self, socket) -> None:
        self._socket = socket
        self._CMD = f'"Qx_Allg_Socket_T13_{socket}"'

    def __get__(self, instance, owner) -> Self | bool:
        if instance is None:
            return self  # pragma: no cover

        value = bool(instance.read(self._CMD))
        state = "ON" if value else "OFF"
        logger.info(f"T13 Power Socket {self._socket} is {state}")
        return value

    def __set__(self, instance, state) -> None:
        validate_bool("state", state, logger)
        instance.write(self._CMD, state)
        state_str = "ON" if state else "OFF"
        logger.info(f"T13 Power Socket {self._socket} is switched {state_str}")


@unique
class _Safety(ValueEnum):
    """
    NodeID strings for the basic safety circuit status
    """

    # Status is a read-only integer containing the state number of the
    # BaseCube-internal state machine. The values correspond to the numbers in
    # :class:`SafetyStatus`.
    STATUS = '"DB_Safety_Circuit"."si_safe_status"'


class SafetyStatus(IntEnum):
    """
    Safety status values that are possible states returned from
    :meth:`hvl_ccb.dev.cube.base.BaseCube.status`. These
    values correspond to the states of the BaseCube's safety circuit statemachine.
    """

    # System is initializing or booting.
    INITIALIZING = 0

    # System is safe, lamps are green and some safety elements are not in place such
    # that it cannot be switched to 'RED_READY' currently.
    GREEN_NOT_READY = 1

    # System is safe and all safety elements are in place to be able to switch to
    # *ready* 'RED_READY'.
    GREEN_READY = 2

    # System is locked in red state and *ready* to go to *operate* ('RED_OPERATE') mode.
    RED_READY = 3

    # System is locked in red state and in *operate* mode, i.e. high voltage on.
    RED_OPERATE = 4

    # Fast turn off triggered and switched off the system. Reset Breakdown Detection
    # to go back to a normal state.
    QUICK_STOP = 5

    # System is in error mode.
    ERROR = 6


class _SafetyStatusTransition(Enum, init="source target command"):  # type: ignore[call-arg]
    """
    NodeID strings for the transition between "ready" and "operate" and
    the corresponding source and target states.
    """

    #: Writable boolean for switching to Red Ready (locked, HV off) state.
    SWITCH_TO_READY = (
        SafetyStatus.GREEN_READY,
        SafetyStatus.RED_READY,
        '"DB_Safety_Circuit"."sx_safe_switch_to_ready"',
    )
    #: Writable boolean for switching to Red Operate (locket, HV on) state.
    SWITCH_TO_OPERATE = (
        SafetyStatus.RED_READY,
        SafetyStatus.RED_OPERATE,
        '"DB_Safety_Circuit"."sx_safe_switch_to_operate"',
    )


STOP_SAFETY_STATUSES: tuple[SafetyStatus, ...] = (
    cast("SafetyStatus", SafetyStatus.GREEN_NOT_READY),
    cast("SafetyStatus", SafetyStatus.GREEN_READY),
)
"""
BaseCube's safety statuses required to close the connection to the device.
"""


@unique
class _Power(ValueEnum):
    """
    Variable NodeID strings concerning power data.
    """

    # Primary voltage in volts, measured by the power inverter at its output.
    # (read-only)
    VOLTAGE_PRIMARY = '"DB_Datamanagement"."si_output_SC_voltage"'

    # Primary current in ampere, measured by the power inverter. (read-only)
    CURRENT_PRIMARY = '"DB_Datamanagement"."si_output_SC_current"'

    # Power setup that is configured using the BaseCube HMI. The value corresponds to
    # the ones in :class:`PowerSetup`. (read-only)
    SETUP = '"DB_Safety_Circuit"."si_power_setup"'

    # Voltage slope in V/s or kV/s (depends on Power Setup).
    VOLTAGE_SLOPE = '"DB_Powercontrol"."si_set_dUdT"'

    # Target voltage setpoint in V or kV (depends on Power Setup).
    VOLTAGE_TARGET = '"DB_Powercontrol"."si_set_voltage"'

    # Maximum voltage allowed by the current experimental setup in V or kV
    # (depends on Power Setup).. (read-only)
    VOLTAGE_MAX = '"DB_Powercontrol"."si_voltage_limit_panel"'

    # Power inverter output frequency. (read-only)
    FREQUENCY = '"DB_Datamanagement"."si_converter_frequency"'

    # Polarity of the output if a DC PowerSetup (7 or 8) is used.
    # Returns True if positive
    POLARITY = '"DB_Powercontrol"."sx_set_polarity"'

    # actual measured output voltage in V or kV (depends on Power Setup).
    VOLTAGE_ACTUAL = '"DB_Measurements"."si_actual_voltage"'


class Polarity(IntEnum):
    NEGATIVE = 0
    POSITIVE = 1


class PowerSetup(IntEnum, init="value slope_min slope_max scale unit"):  # type: ignore[call-arg]
    """
    Possible power setups corresponding to the value of variable :attr:`Power.setup`.
    The values for slope_min are experimentally defined, below these values the slope
    is more like a staircase

    The name of the first argument needs to be 'value', otherwise the IntEnum is
    not working correctly.
    """

    # No safety switches, uses only safety components (doors, fence, earthing...)
    # without any power source.
    NO_SOURCE = 0, 0, 0, 1, ""

    # For PICube: External power supply fed through blue 25A power plug input using
    # isolation transformer and safety switches of the PICube
    # For BaseCube: Use of an external safety switch attached to the BaseCube.
    EXTERNAL_SOURCE = 1, 0, 0, 1, ""

    # AC setup with one MWB transformer set to 50 kV maximum voltage.
    AC_50KV = 2, 100, 6000, 1e-3, "kV"

    # AC setup with one MWB transformer set to 100 kV maximum voltage.
    AC_100KV = 3, 100, 15000, 1e-3, "kV"

    # AC setup with two MWB transformers, one configured to a output voltage
    # of 100 kV and the other to 50 kV, resulting in a total maximum voltage of 150 kV.
    AC_150KV = 4, 200, 15000, 1e-3, "kV"

    # AC setup with two MWB transformers both configured to a output voltage
    # of 100 kV, resulting in a total maximum voltage of 200kV.
    AC_200KV = 5, 200, 15000, 1e-3, "kV"

    # Direct control of the internal power inverter, controlling of the primary voltage
    # output of the PICube itself. The maximum voltage at the output of the PICube
    # is 200 V. No feedback loop with a measurement transformer is used.
    POWER_INVERTER_220V = 6, 0.2, 15, 1, "V"

    # DC setup with one AC transformer configured to 100 kV and a rectifier circuit.
    # The maximum DC voltage is 140 kV.
    DC_140KV = 7, 300, 15000, 1e-3, "kV"

    # DC setup with one AC transformer configured to 100 kV and a Greinacher
    # voltage doubler circuit.
    # OR a DC setup with two AC transformers both configured to 100 kV and a rectifier
    # circuit. Both setup are resulting in DC voltage of 280 kV.
    DC_280KV = 8, 300, 15000, 1e-3, "kV"

    # Impulse setup with one AC transformer configured to 100 kV and a rectifier
    # circuit, which results in a maximum DC voltage of 140 kV. The impulse is
    # triggered with a spark gap.
    IMPULSE_140KV = 9, 300, 15000, 1e-3, "kV"


DC_POWER_SETUPS: tuple[PowerSetup, ...] = (
    cast("PowerSetup", PowerSetup.DC_140KV),
    cast("PowerSetup", PowerSetup.DC_280KV),
)

AC_POWER_SETUPS: tuple[PowerSetup, ...] = (
    cast("PowerSetup", PowerSetup.AC_50KV),
    cast("PowerSetup", PowerSetup.AC_100KV),
    cast("PowerSetup", PowerSetup.AC_150KV),
    cast("PowerSetup", PowerSetup.AC_200KV),
)


class _MeasurementChannel:
    """
    Measurement Channel with properties for the value and the ratio.
    """

    def __init__(self, handle, number: int, input_noise: Number) -> None:
        self._handle: BaseCube = handle
        self._number: int = number
        self._input_noise: Number = input_noise
        self._CMD_SCALE: str = f'"DB_Measurements"."sx_volts_input_{number}"'
        self._CMD_VOLTAGE: str = f'"DB_Measurements"."si_scaled_Voltage_Input_{number}"'
        self._CMD_RATIO: str = f'"DB_Measurements"."si_Divider_Ratio_{number}"'

    @property
    def voltage(self) -> float:
        """
        Measured voltage of the measurement channel.

        :return: in V
        """
        value = float(self._handle.read(self._CMD_VOLTAGE))
        scale_unit = self._handle.read(self._CMD_SCALE)
        if scale_unit:
            value_return = value * 1e3
            unit = "kV"
        else:
            value_return = value
            unit = "V"
        logger.info(
            f"Measurement Voltage of Channel {self._number} is {value:_.2f} {unit}"
        )
        return value_return

    @property
    def ratio(self) -> float:
        """
        Set ratio for the measurement channel.

        :return: in 1
        """
        value = float(self._handle.read(self._CMD_RATIO))
        logger.info(f"Measurement Ratio of Channel {self._number} is {value}")
        return value

    @property
    def noise_level(self) -> Number:
        return self._input_noise


@unique
class _Errors(ValueEnum):
    """
    Variable NodeID strings for information regarding error, warning and message
    handling.
    """

    #: Boolean read-only variable telling if a message is active.
    MESSAGE = '"DB_Message_Buffer"."Info_active"'

    #: Boolean read-only variable telling if a warning is active.
    WARNING = '"DB_Message_Buffer"."Warning_active"'

    #: Boolean read-only variable telling if a stop is active.
    STOP = '"DB_Message_Buffer"."Stop_active"'

    #: Writable boolean for the error quit button.
    QUIT = '"DB_Message_Buffer"."Reset_button"'


class DoorStatus(IntEnum):
    """
    Possible status values for doors.
    """

    #: not enabled in BaseCube HMI setup, this door is not supervised.
    INACTIVE = 0

    #: Door is open.
    OPEN = 1

    #: Door is closed, but not locked.
    CLOSED = 2

    #: Door is closed and locked (safe state).
    LOCKED = 3

    #: Door has an error or was opened in locked state (either with emergency stop or
    #: from the inside).
    ERROR = 4


class _Door(_BaseGetDescriptor):
    """
    Get the status of a safety fence door. See :class:`constants.DoorStatus` for
    possible returned door statuses.
    """

    def __init__(self, number, name) -> None:
        super().__init__(number, name)
        self._CMD = f'"DB_Safety_Circuit"."Door_{number}"."si_HMI_status"'

    def __get__(self, instance, owner) -> Self | DoorStatus:
        """
        :return: the door status
        """
        if instance is None:
            return self  # pragma: no cover

        value = DoorStatus(instance.read(self._CMD))
        logger.info(f"Door {self.number} is {value.name}")
        return value


class EarthingRodStatus(IntEnum):
    """
    Possible status values for earthing rods.
    """

    #: earthing rod is somewhere in the experiment
    #: and blocks the start of the experiment
    EXPERIMENT_BLOCKED = 0

    #: earthing rod is hanging next to the door, experiment is ready to operate
    EXPERIMENT_READY = 1


class _EarthingRod(_BaseGetDescriptor):
    """
    Get the status of a earthing rod. See :class:`constants.EarthingRodStatus` for
    possible returned earthing rod statuses.
    """

    def __init__(self, number, name) -> None:
        super().__init__(number, name)
        self._CMD = f'"DB_Safety_Circuit"."Door_{number}"."Ix_earthingrod"'

    def __get__(self, instance, owner) -> Self | EarthingRodStatus:
        """
        :return: the earthing rod status
        """
        if instance is None:
            return self  # pragma: no cover

        value = EarthingRodStatus(instance.read(self._CMD))
        status = "NOT " if value == EarthingRodStatus.EXPERIMENT_READY else ""
        logger.info(f"Earthing Rod {self.number} is {status}blocking the Experiment")
        return value


class _PrefixedNumbersEnumBase(ValueEnum):
    """
    Base class for enums with "{prefix}{n}" instance names, where n=1..N.
    """

    @classmethod
    def range(cls) -> Sequence[int]:
        """
        Integer range of all channels.

        :return: sequence of channel numbers
        """
        return range(1, len(cast("Sized", cls)) + 1)

    @classmethod
    def _validate_number(cls, number: int) -> None:
        """
        Validate enum instance number.

        :param number: the enum instance number (1..N)
        :raises ValueError: when enum instance number is not in 1..N range
        """
        if number not in cls.range():
            msg = f"{cls._prefix()} number must be one of {list(cls.range())}"
            raise ValueError(msg)

    @classmethod
    def _prefix(cls) -> str:
        """
        Enum instances name prefix: "{prefix}{n}"

        :return: enum instances prefix string
        """
        msg = "Implement in subclass"
        raise NotImplementedError(msg)  # pragma: no cover

    @property
    def number(self) -> int:
        """
        Get corresponding enum instance number.

        :return: enum instance number (1..N)
        """
        return self.name.removeprefix(self._prefix())

    # no type return as it would be a arguably too complex/obscure;
    # cf. https://github.com/python/typing/issues/58#issuecomment-326240794
    @classmethod
    def get(cls, number: int) -> _PrefixedNumbersEnumBase:
        """
        Get the enum instance for a given number.

        :param number: the instance number (1..N)
        :return: the enum instance for the given number.
        :raises ValueError: when instance number is not in the 1..N range
        """
        cls._validate_number(number)
        return getattr(cls, f"{cls._prefix()}{number}")


class _OpcControl(ValueEnum):
    """
    Variable NodeID strings for supervision of the OPC connection from the
    controlling workstation to the BaseCube.
    """

    # writable boolean to enable OPC remote control and display a message window on
    # the BaseCube HMI.
    ACTIVE = '"DB_OPC_Connection"."sx_OPC_active"'
    LIVE = '"DB_OPC_Connection"."sx_OPC_lifebit"'
    TIME = '"DB_OPC_Connection"."st_system_time"'


NUMBER_OF_LINES = 16


class _LineEnumBase(_PrefixedNumbersEnumBase):
    """
    Base class for enums with "input_{n}" instance names, where n=1..N.
    """

    @classmethod
    def _prefix(cls) -> str:
        return "LINE_"

    @classmethod
    def line(cls, number: int) -> _LineEnumBase:
        """
        Get the enum instance for a given line number.

        :param number: the line number (1..M)
        :return: the enum instance for the given line number.
        :raises ValueError: when line number is not in the 1..N range
        """
        return cls.get(number)


MessageBoard = unique(
    _LineEnumBase(
        "MessageBoard",
        {
            f"{_LineEnumBase._prefix()}{n}": f'"DB_OPC_Connection"."Is_status_Line_{n}"'
            for n in range(1, NUMBER_OF_LINES)
        },
    )
)
"""
Variable NodeID strings for message board lines.
"""



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Device class for Pfeiffer TPG controllers.

The Pfeiffer TPG control units are used to control Pfeiffer Compact Gauges.
Models: TPG 251 A, TPG 252 A, TPG 256A, TPG 261, TPG 262, TPG 361, TPG 362 and TPG 366.

Manufacturer homepage:
https://www.pfeiffer-vacuum.com/en/products/measurement-analysis/
measurement/activeline/controllers/
"""

from .pfeiffer_tpg import (  # noqa: F401
    PfeifferTPG,
    PfeifferTPGConfig,
    PfeifferTPGError,
    PfeifferTPGSerialCommunication,
    PfeifferTPGSerialCommunicationConfig,
)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
VISA masked communication protocol module.
"""

from collections import defaultdict
from queue import Queue

from hvl_ccb.comm.visa import VisaCommunication
from hvl_ccb.utils.typing import Number


class MaskedVisaCommunication(VisaCommunication):
    """
    Masked version of VisaCommunication to simulate messages.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._read_buffer = defaultdict(Queue)
        self._write_buffer = Queue()

        self.stb = 0

    def open(self) -> None:
        pass

    def close(self) -> None:
        pass

    def write(self, text: str) -> None:
        if text == "*OPC":
            self.put_name("*ESR?", "1")

        self._write_buffer.put(text)

    def write_multiple(self, *commands: str) -> None:
        if "*OPC" in commands:
            # operation complete request set, immediately reply
            self.put_name("*ESR?", "1")

        for command in commands:
            self._write_buffer.put(command)

    def query(
        self,
        command: str,
        _n_attempts_max: int | None = None,
        _attempt_interval_sec: Number | None = None,
    ) -> str | None:
        return (
            self._read_buffer[command].get()
            if not self._read_buffer[command].empty()
            else "0"
        )

    def query_multiple(self, *commands: str) -> str | tuple[str, ...]:
        out = [
            (
                self._read_buffer[command].get()
                if not self._read_buffer[command].empty()
                else "0"
            )
            for command in commands
        ]

        return out[0] if len(out) == 1 else tuple(out)

    def spoll(self) -> int:
        return self.stb

    def put_name(self, command: str, string: str) -> None:
        self._read_buffer[command].put(string)

    def get_written(self) -> str | None:
        return self._write_buffer.get() if not self._write_buffer.empty() else None



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#

"""
Sensors that are used by the devices implemented in the CCB
"""

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import cast

import numpy as np
import numpy.typing as npt

from hvl_ccb.utils.typing import ConvertableTypes
from hvl_ccb.utils.validation import validate_number

from .unit import Temperature
from .utils import GetAttr, SetAttr, convert_value_to_str, preserve_type

logger = logging.getLogger(__name__)


class Sensor(ABC):
    @abstractmethod
    @preserve_type
    def convert(self, value: ConvertableTypes) -> ConvertableTypes:
        pass  # pragma: no cover


@dataclass
class LEM4000S(Sensor):
    CONVERSION = GetAttr(5000, "CONVERSION")
    shunt = SetAttr(1.2, "shunt", limits=(1e-6, None), validator=validate_number)
    calibration_factor = SetAttr(
        1,
        "calibration_factor",
        limits=(0.9, 1.1),
        absolut=True,
        validator=validate_number,
    )

    @preserve_type
    def convert(self, value: ConvertableTypes) -> ConvertableTypes:
        value = cast("npt.NDArray", value)
        conversion = cast("int", self.CONVERSION) * cast(
            "float", self.calibration_factor
        )
        value_ret = value / self.shunt * conversion
        value_str, value_ret_str = convert_value_to_str(value, value_ret)
        logger.info(
            f"LEM400S: Convert a secondary current of {value_str} A to a "
            f"primary current of {value_ret_str} A."
        )
        return value_ret


@dataclass
class LMT70A(Sensor):
    """
    Converts the output voltage (V) to the measured temperature (default °C)
    when using a TI Precision Analog Temperature Sensor LMT70(A)

    """

    temperature_unit = SetAttr(
        Temperature.CELSIUS, "temperature_unit", limits=None, validator=Temperature
    )

    # look up table from datasheet
    # first column: temperature in degree celsius
    # second column: voltage in volt
    # https://www.ti.com/lit/ds/symlink/lmt70a.pdf?ts=1631590373860
    LUT = GetAttr(
        np.array(
            [
                [-55.0, 1.375219],
                [-50.0, 1.350441],
                [-40.0, 1.300593],
                [-30.0, 1.250398],
                [-20.0, 1.199884],
                [-10.0, 1.14907],
                [0.0, 1.097987],
                [10.0, 1.046647],
                [20.0, 0.99505],
                [30.0, 0.943227],
                [40.0, 0.891178],
                [50.0, 0.838882],
                [60.0, 0.78636],
                [70.0, 0.733608],
                [80.0, 0.680654],
                [90.0, 0.62749],
                [100.0, 0.574117],
                [110.0, 0.520551],
                [120.0, 0.46676],
                [130.0, 0.412739],
                [140.0, 0.358164],
                [150.0, 0.302785],
            ]
        ),
        "LUT",
    )

    @preserve_type
    def convert(self, value: ConvertableTypes) -> ConvertableTypes:
        """
        NaN is returned for values that are not covered by the look up table
        :param value: output voltage of the sensor.
        :raise TypeError: for non convertable data types
        :return: measured temperature (default °C)
        """
        # cast necessary because of wrapper, which changes the type of value
        value = cast("npt.NDArray", value)
        try:
            validate_number(
                "value",
                value,
                (
                    cast("npt.NDArray", self.LUT)[-1, 1],
                    cast("npt.NDArray", self.LUT)[0, 1],
                ),
                logger=logger,
            )
        except ValueError:
            mask = np.any(
                [
                    value < cast("npt.NDArray", self.LUT)[-1, 1],
                    value > cast("npt.NDArray", self.LUT)[0, 1],
                ],
                axis=0,
            )
            value[mask] = np.nan
        logger.info(
            "LMT70A: Use linear interpolation of lookup table provided "
            "in datasheet of LMT70A"
        )
        value_ret = np.interp(
            value,
            cast("npt.NDArray", self.LUT)[::-1, 1],
            cast("npt.NDArray", self.LUT)[::-1, 0],
        )
        value_ret = Temperature.convert(
            value_ret, source=Temperature.CELSIUS, target=self.temperature_unit
        )
        value_str, value_ret_str = convert_value_to_str(value, value_ret)
        logger.info(
            f"LMT70A: Convert a voltage of {value_str} V to "
            f"a temperature of {value_ret_str} "
            f"{cast('Temperature', self.temperature_unit).name}."
        )
        return value_ret



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Switches of the different "Cubes".
"""

import logging
from abc import ABC
from typing import TYPE_CHECKING

from aenum import IntEnum

from hvl_ccb.utils.enum import BoolEnum

from .constants import SafetyStatus
from .errors import SwitchOperationError

if TYPE_CHECKING:
    from . import BaseCube  # pragma: no cover

logger = logging.getLogger(__name__)


class SwitchStatus(IntEnum):
    """
    Status of a switch. These are the possible values in the status integer
    e.g. in :attr:`_Switch.status`.
    """

    # Switch is deselected and not enabled in safety circuit. To get out of
    # this state, the earthing has to be enabled in the BaseCube HMI setup.
    INACTIVE = 0

    # Earthing is closed (safe).
    CLOSED = 1

    # Earthing is open (not safe).
    OPEN = 2

    # Earthing is in error, e.g. when the stick did not close correctly or could not
    # open.
    ERROR = 3


class SwitchOperatingStatus(IntEnum):
    """
    Operating Status of a switch. Switch can be in auto or manual mode.
    """

    AUTO = 0
    MANUAL = 1


class SwitchOperation(BoolEnum):
    """
    Operation of a switch in manual operating mode. Can be closed or opened.
    """

    OPEN = False
    CLOSE = True


class _Switch(ABC):  # noqa: B024
    """
    Switch with status, operating status (manual and auto) and manual operate.
    """

    # TODO: re-write `_Switch` to not be an `ABC`  # noqa: FIX002
    # https://gitlab.com/ethz_hvl/hvl_ccb/-/issues/381

    _ERROR_CLS = SwitchOperationError
    _SWITCHABLE_AT_STATES: tuple[SafetyStatus] | None = None

    def __init__(self, handle, device_name: str) -> None:
        self._handle: BaseCube = handle
        self._device_name: str = device_name
        self._CMD_STATUS: str | None = None
        self._CMD_OPERATING_STATUS: str | None = None
        self._CMD_MANUAL: str | None = None

    @property
    def status(self) -> SwitchStatus:
        """
        Position status of a switch.

        :return: Status of the switch.
        """
        if self._CMD_STATUS is None:
            msg = (
                f"Tried to query position status of {self._device_name}, "
                f"but {self._device_name} does not have a position status feedback."
            )
            logger.error(msg)
            raise self._ERROR_CLS(msg)

        value = SwitchStatus(self._handle.read(self._CMD_STATUS))
        logger.info(f"Status of {self._device_name} is {value.name}")
        return value

    @property
    def operating_status(self) -> SwitchOperatingStatus:
        """
        Switch operating status, if 'manual' the stick can be controlled by the
        user.

        :return: Switch operating status, can be either auto or manual
        """
        if self._CMD_OPERATING_STATUS is None:
            msg = (
                f"Tried to query operating status of {self._device_name}, "
                f"but {self._device_name} does not have an operating status."
            )
            logger.error(msg)
            raise self._ERROR_CLS(msg)

        value = SwitchOperatingStatus(self._handle.read(self._CMD_OPERATING_STATUS))
        logger.info(f"Operating Status of {self._device_name} is {value.name}")
        return value

    def operation_conditions_fulfilled(self) -> None:
        """
        Method to be called before an operation is performed. It will raise an
        exception if not
        """
        if (
            self._SWITCHABLE_AT_STATES
            and self._handle.status not in self._SWITCHABLE_AT_STATES
        ):
            switchable_states = " or ".join(
                [f'"{_}"' for _ in self._SWITCHABLE_AT_STATES]
            )
            msg = (
                f"Cube needs to be in state {switchable_states} "
                f"to operate {self._device_name} manually, "
                f'but is in "{self._handle.status.name}".'
            )
            logger.error(msg)
            raise self._ERROR_CLS(msg)

        if self.operating_status != SwitchOperatingStatus.MANUAL:
            msg = (
                f"Operation of the {self._device_name} is not possible, "
                "as the feature is not activated in the Cube Setup."
            )
            logger.error(msg)
            raise self._ERROR_CLS(msg)

    @property
    def operate(self) -> SwitchOperation:
        """
        Operation of a switch which has a manual operating status.

        :return: Switch operation setting, can be open or close
        """
        if self._CMD_MANUAL is None:
            msg = (
                f"Tried to query manual operation setting of {self._device_name}, "
                f"but {self._device_name} cannot be operated manually."
            )
            logger.error(msg)
            raise self._ERROR_CLS(msg)

        value = SwitchOperation(self._handle.read(self._CMD_MANUAL))
        logger.info(f"Manual operation status of {self._device_name} is {value}")
        return value

    @operate.setter
    def operate(self, operation: SwitchOperation) -> None:
        """
        Operation of a switch which has a manual operating status.

        :param operation: switch operation setting (close or open)
        :raises SwitchOperationError: when the operation conditions of the switch
        are not fulfilled, or if the switch is not allowed to be manually operated
        """
        self.operation_conditions_fulfilled()

        if self._CMD_MANUAL is None:
            msg = (  # pragma: no cover
                f"Tried to operate {self._device_name}, "
                f"but {self._device_name} cannot"
                "be operated manually."
            )
            logger.error(msg)  # pragma: no cover
            raise self._ERROR_CLS(msg)  # pragma: no cover

        operation = SwitchOperation(operation)
        self._handle.write(self._CMD_MANUAL, operation)
        logger.info(f"{self._device_name} is set to {operation}")



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Testing PFA-2 Filter.
"""

import logging

import pytest

from hvl_ccb.dev.pfa2_filter.base import (
    N_CHANNELS,
    Pfa2FilterChannelCoupling,
    Pfa2FilterChannelMode,
    Pfa2FilterCommunicationError,
    Pfa2FilterError,
    Pfa2FilterLPFMode,
    Pfa2FilterOverloadMode,
    Pfa2FilterPostGain,
    Pfa2FilterPreGain,
    Pfa2FilterSerialCommunicationConfig,
)
from hvl_ccb.dev.pfa2_filter.device import (
    Pfa2Filter,
    Pfa2FilterConfig,
)
from masked_comm.serial import Pfa2FilterLoopSerialCommunication

logging.basicConfig(level=logging.ERROR)


@pytest.fixture(scope="module")
def com_config():
    return {
        "port": "loop://?logging=debug",
        "baudrate": 57600,
        "timeout": 0.2,
        "default_n_attempts_read_text_nonempty": 2,
        "wait_sec_read_text_nonempty": 0.1,
    }


@pytest.fixture(scope="module")
def dev_config():
    return {
        "device_option_hpf": True,
        "coupling_init": Pfa2FilterChannelCoupling.DC,
        "mode_init": Pfa2FilterChannelMode.OPERATE,
        "pregain_init": 1,
        "postgain_init": 1,
        "lpf_freq_init": 127750,
        "lpf_mode_init": Pfa2FilterLPFMode.FLAT,
        "hpf_state_init": True,
        "hpf_freq_init": 1,
        "overload_limit_init": 10,
        "overload_mode_init": Pfa2FilterOverloadMode.CONTINUOUS,
    }


class ConcretePfa2Filter(Pfa2Filter):
    def start(self) -> None:
        super().start()


@pytest.fixture
def started_pfa2filter_device(com_config, dev_config):
    serial_port = Pfa2FilterLoopSerialCommunication(com_config)
    serial_port.open()
    for _ in range(N_CHANNELS):
        serial_port.put_text("DC")
        serial_port.put_text("OPERATE")
        serial_port.put_text("1")
        serial_port.put_text("1")
        serial_port.put_text("127750")
        serial_port.put_text("FLAT")
        serial_port.put_text("10")
        serial_port.put_text("CONTINUOUS")
        serial_port.put_text("ON")
        serial_port.put_text("1")

    with Pfa2Filter(serial_port, dev_config) as pfa2filter:
        while serial_port.get_written() is not None:
            pass
        yield serial_port, pfa2filter


@pytest.fixture
def start_pfa2filter_device(com_config, dev_config):
    def _start_pfa2filter_device():
        serial_port = Pfa2FilterLoopSerialCommunication(com_config)
        serial_port.open()

        def started_pfa2filter():
            with Pfa2Filter(serial_port, dev_config) as pfa2filter:
                while serial_port.get_written() is not None:
                    pass
                yield serial_port, pfa2filter

        return started_pfa2filter()

    return _start_pfa2filter_device


def test_com_config(com_config) -> None:
    config = Pfa2FilterSerialCommunicationConfig(**com_config)
    for key, value in com_config.items():
        assert getattr(config, key) == value


def test_dev_config(dev_config) -> None:
    # currently there are no non-default config values
    Pfa2FilterConfig()

    config = Pfa2FilterConfig(**dev_config)
    for key, value in dev_config.items():
        assert getattr(config, key) == value


@pytest.mark.parametrize(
    "wrong_config_dict",
    [
        {"coupling_init": True},
        {"mode_init": 2},
        {"pregain_init": "test"},
        {"postgain_init": "test"},
        {"lpf_freq_init": "test"},
        {"hpf_state_init": "test"},
        {"hpf_freq_init": "test"},
        {"overload_limit_init": "test"},
        {"overload_mode_init": 2},
    ],
)
def test_invalid_config_dict(dev_config, wrong_config_dict) -> None:
    invalid_config = dict(dev_config)
    invalid_config.update(wrong_config_dict)

    with pytest.raises(TypeError):
        Pfa2FilterConfig(**invalid_config)


def test_pfa2filter_instantiation(com_config, dev_config) -> None:
    pfa2filter_1 = ConcretePfa2Filter(com_config)
    assert pfa2filter_1 is not None

    pfa2filter_2 = ConcretePfa2Filter(com_config, dev_config)
    assert pfa2filter_2 is not None

    assert pfa2filter_2.__str__() == "PFA-2 Filter"


def test_pfa2filter_stop(started_pfa2filter_device) -> None:
    com, pfa2filter = started_pfa2filter_device
    pfa2filter.stop()

    assert not com.get_written()

    # check that a second stop() works
    pfa2filter.stop()
    assert not com.get_written()


def test_pfa2filter_channel_coupling(started_pfa2filter_device) -> None:
    com, pfa2filter = started_pfa2filter_device

    com.put_text("DC")
    assert pfa2filter.channel_dict[0].coupling == Pfa2FilterChannelCoupling.DC
    assert com.get_written() == "0:COUPLING?"

    com.put_text("AC")
    pfa2filter.channel_dict[1].coupling = Pfa2FilterChannelCoupling.AC
    assert com.get_written() == "1:COUPLING=AC"

    with pytest.raises(Pfa2FilterCommunicationError):
        assert pfa2filter.channel_dict[0].coupling == Pfa2FilterChannelCoupling.DC

    with pytest.raises(TypeError):
        pfa2filter.channel_dict[0].coupling = 1

    with pytest.raises(ValueError):
        pfa2filter.channel_dict[0].coupling = "tea with milk"

    with pytest.raises(Pfa2FilterCommunicationError):
        pfa2filter.channel_dict[1].coupling = Pfa2FilterChannelCoupling.AC


def test_pfa2filter_channel_input_mode(
    started_pfa2filter_device,
) -> None:
    com, pfa2filter = started_pfa2filter_device

    com.put_text("OPERATE")
    assert pfa2filter.channel_dict[0].input_mode == Pfa2FilterChannelMode.OPERATE
    assert com.get_written() == "0:MODE?"

    com.put_text("SHORT")
    pfa2filter.channel_dict[1].input_mode = Pfa2FilterChannelMode.SHORTED
    assert com.get_written() == "1:MODE=SHORT"

    with pytest.raises(Pfa2FilterCommunicationError):
        assert pfa2filter.channel_dict[0].input_mode == Pfa2FilterChannelMode.OPERATE

    with pytest.raises(TypeError):
        pfa2filter.channel_dict[0].input_mode = 1

    with pytest.raises(ValueError):
        pfa2filter.channel_dict[0].input_mode = "tea with milk"

    with pytest.raises(Pfa2FilterCommunicationError):
        pfa2filter.channel_dict[0].input_mode = Pfa2FilterChannelMode.OPERATE


def test_pfa2filter_channel_lpf_mode(
    started_pfa2filter_device,
) -> None:
    com, pfa2filter = started_pfa2filter_device

    com.put_text("FLAT")
    assert pfa2filter.channel_dict[0].lpf_mode == Pfa2FilterLPFMode.FLAT
    assert com.get_written() == "0:LPFILTTYPE?"

    com.put_text("PULSE")
    assert pfa2filter.channel_dict[1].lpf_mode == Pfa2FilterLPFMode.PULSE

    with pytest.raises(Pfa2FilterCommunicationError):
        assert pfa2filter.channel_dict[0].lpf_mode == Pfa2FilterLPFMode.FLAT

    with pytest.raises(TypeError):
        pfa2filter.channel_dict[0].lpf_mode = True

    with pytest.raises(ValueError):
        pfa2filter.channel_dict[0].lpf_mode = "this is a test to fail"

    with pytest.raises(Pfa2FilterCommunicationError):
        pfa2filter.channel_dict[0].lpf_mode = Pfa2FilterLPFMode.FLAT


def test_pfa2filter_channel_lpf_freq(
    started_pfa2filter_device,
) -> None:
    com, pfa2filter = started_pfa2filter_device

    com.put_text("127750")
    assert pfa2filter.channel_dict[0].lpf_freq == 127750
    assert com.get_written() == "0:LPFC?"

    com.put_text("10")
    pfa2filter.channel_dict[1].lpf_freq = 6
    assert com.get_written() == "1:LPFC=10"

    com.put_text("2750")
    pfa2filter.channel_dict[1].lpf_freq = 2700
    assert com.get_written() == "1:LPFC=2750"

    com.put_text("127500")
    pfa2filter.channel_dict[1].lpf_freq = 127700
    assert com.get_written() == "1:LPFC=127750"

    with pytest.raises(Pfa2FilterCommunicationError):
        assert pfa2filter.channel_dict[0].lpf_freq == 127750

    with pytest.raises(TypeError):
        pfa2filter.channel_dict[0].lpf_freq = "eggs and bacon"

    with pytest.raises(ValueError):
        pfa2filter.channel_dict[0].lpf_freq = 200000

    with pytest.raises(Pfa2FilterCommunicationError):
        pfa2filter.channel_dict[1].lpf_freq = 127750


def test_pfa2filter_channel_hpf_freq(
    started_pfa2filter_device,
) -> None:
    com, pfa2filter = started_pfa2filter_device

    com.put_text("125000")
    assert pfa2filter.channel_dict[0].hpf_freq == 125000
    assert com.get_written() == "0:HPFC?"

    com.put_text("200")
    pfa2filter.channel_dict[1].hpf_freq = 200
    assert com.get_written() == "1:HPFC=200"

    com.put_text("255")
    pfa2filter.channel_dict[1].hpf_freq = 275
    assert com.get_written() == "1:HPFC=255"

    com.put_text("10000")
    pfa2filter.channel_dict[1].hpf_freq = 10001
    assert com.get_written() == "1:HPFC=10000"

    com.put_text("12750")
    pfa2filter.channel_dict[1].hpf_freq = 12755
    assert com.get_written() == "1:HPFC=12750"

    com.put_text("127000")
    pfa2filter.channel_dict[1].hpf_freq = 127200
    assert com.get_written() == "1:HPFC=127000"

    with pytest.raises(Pfa2FilterCommunicationError):
        assert pfa2filter.channel_dict[0].hpf_freq == 125000

    with pytest.raises(TypeError):
        pfa2filter.channel_dict[0].hpf_freq = "gin and tonic"

    with pytest.raises(ValueError):
        pfa2filter.channel_dict[0].hpf_freq = 200000

    with pytest.raises(Pfa2FilterCommunicationError):
        pfa2filter.channel_dict[0].hpf_freq = 12755

    pfa2filter.channel_dict[1]._option_hpf = False
    with pytest.raises(Pfa2FilterError):
        assert pfa2filter.channel_dict[1].hpf_freq == 12755

    with pytest.raises(Pfa2FilterError):
        pfa2filter.channel_dict[1].hpf_freq = 12755


def test_pfa2filter_channel_hpf_state(
    started_pfa2filter_device,
) -> None:
    com, pfa2filter = started_pfa2filter_device

    com.put_text("OFF")
    assert pfa2filter.channel_dict[1].hpf_state is False
    assert com.get_written() == "1:HPFILT?"

    com.put_text("ON")
    pfa2filter.channel_dict[0].hpf_state = True
    assert com.get_written() == "0:HPFILT=ON"

    with pytest.raises(Pfa2FilterCommunicationError):
        assert pfa2filter.channel_dict[0].hpf_state is False

    with pytest.raises(TypeError):
        pfa2filter.channel_dict[0].hpf_state = "not a boolean"

    with pytest.raises(Pfa2FilterCommunicationError):
        pfa2filter.channel_dict[0].hpf_state = True

    pfa2filter.channel_dict[1]._option_hpf = False
    with pytest.raises(Pfa2FilterError):
        assert pfa2filter.channel_dict[1].hpf_state is False

    with pytest.raises(Pfa2FilterError):
        pfa2filter.channel_dict[1].hpf_state = False


def test_pfa2filter_channel_pregain(
    started_pfa2filter_device,
) -> None:
    com, pfa2filter = started_pfa2filter_device

    com.put_text("1.0")
    assert pfa2filter.channel_dict[0].pregain == Pfa2FilterPreGain.ONE
    assert com.get_written() == "0:PREGAIN?"

    com.put_text("4")
    pfa2filter.channel_dict[1].pregain = Pfa2FilterPreGain.FOUR
    assert com.get_written() == "1:PREGAIN=4.0"

    com.put_text("4")
    pfa2filter.channel_dict[1].pregain = 5.5
    assert com.get_written() == "1:PREGAIN=4.0"

    com.put_text("128")
    pfa2filter.channel_dict[1].pregain = 256
    assert com.get_written() == "1:PREGAIN=128.0"

    with pytest.raises(Pfa2FilterCommunicationError):
        assert pfa2filter.channel_dict[0].pregain == Pfa2FilterPreGain.ONE

    with pytest.raises(TypeError):
        pfa2filter.channel_dict[0].pregain = "tree fiddy"

    with pytest.raises(ValueError):
        pfa2filter.channel_dict[0].pregain = 3.50e-5

    with pytest.raises(Pfa2FilterCommunicationError):
        pfa2filter.channel_dict[0].pregain = Pfa2FilterPreGain.ONE


def test_pfa2filter_channel_postgain(
    started_pfa2filter_device,
) -> None:
    com, pfa2filter = started_pfa2filter_device

    com.put_text("0.5")
    assert pfa2filter.channel_dict[0].postgain == Pfa2FilterPostGain.ONE_HALF
    assert com.get_written() == "0:POSTGAIN?"

    com.put_text("16")
    pfa2filter.channel_dict[1].postgain = Pfa2FilterPostGain.SIXTEEN
    assert com.get_written() == "1:POSTGAIN=16.0"

    com.put_text("16")
    pfa2filter.channel_dict[1].postgain = 100
    assert com.get_written() == "1:POSTGAIN=16.0"

    com.put_text("8")
    pfa2filter.channel_dict[1].postgain = 8.5
    assert com.get_written() == "1:POSTGAIN=8.0"

    with pytest.raises(Pfa2FilterCommunicationError):
        assert pfa2filter.channel_dict[0].postgain == Pfa2FilterPostGain.ONE_HALF

    with pytest.raises(TypeError):
        pfa2filter.channel_dict[0].postgain = "nothing or something"

    with pytest.raises(ValueError):
        pfa2filter.channel_dict[0].postgain = 0.0001

    with pytest.raises(Pfa2FilterCommunicationError):
        pfa2filter.channel_dict[0].postgain = Pfa2FilterPostGain.ONE_HALF


def test_pfa2filter_channel_overload_state(
    started_pfa2filter_device,
) -> None:
    com, pfa2filter = started_pfa2filter_device

    com.put_text("1")
    assert pfa2filter.channel_dict[0].output_overload is True
    assert com.get_written() == "0:OUTOVLD?"

    com.put_text("0")
    assert pfa2filter.channel_dict[1].input_overload is False
    assert com.get_written() == "1:INOVLD?"

    with pytest.raises(Pfa2FilterCommunicationError):
        assert pfa2filter.channel_dict[0].output_overload is False

    with pytest.raises(Pfa2FilterCommunicationError):
        assert pfa2filter.channel_dict[1].input_overload is True


def test_pfa2filter_channel_overload_limit(
    started_pfa2filter_device,
) -> None:
    com, pfa2filter = started_pfa2filter_device

    com.put_text("1.0")
    assert pfa2filter.channel_dict[0].overload_limit == 1.0
    assert com.get_written() == "0:OUTOVLDLIM?"

    com.put_text("4")
    pfa2filter.channel_dict[1].overload_limit = 4.0
    assert com.get_written() == "1:OUTOVLDLIM=4.0"

    with pytest.raises(Pfa2FilterCommunicationError):
        assert pfa2filter.channel_dict[0].overload_limit == 1.0

    with pytest.raises(TypeError):
        pfa2filter.channel_dict[0].overload_limit = "hello there"

    with pytest.raises(ValueError):
        pfa2filter.channel_dict[0].overload_limit = 125

    with pytest.raises(Pfa2FilterCommunicationError):
        pfa2filter.channel_dict[0].overload_limit = 5


def test_pfa2filter_channel_overload_mode(
    started_pfa2filter_device,
) -> None:
    com, pfa2filter = started_pfa2filter_device

    com.put_text("LATCHING")
    assert pfa2filter.channel_dict[0].overload_mode == Pfa2FilterOverloadMode.LATCHING
    assert com.get_written() == "0:OVLDMODE?"

    com.put_text("CONTINUOUS")
    pfa2filter.channel_dict[1].overload_mode = Pfa2FilterOverloadMode.CONTINUOUS
    assert com.get_written() == "1:OVLDMODE=CONTINUOUS"

    with pytest.raises(Pfa2FilterCommunicationError):
        assert (
            pfa2filter.channel_dict[0].overload_mode == Pfa2FilterOverloadMode.LATCHING
        )

    with pytest.raises(TypeError):
        pfa2filter.channel_dict[0].overload_mode = 1

    with pytest.raises(ValueError):
        pfa2filter.channel_dict[0].overload_mode = "you shall not pass"

    with pytest.raises(Pfa2FilterCommunicationError):
        pfa2filter.channel_dict[0].overload_mode = Pfa2FilterOverloadMode.LATCHING


def test_pfa2filter_channel_overload_clear(
    started_pfa2filter_device,
) -> None:
    com, pfa2filter = started_pfa2filter_device

    com.put_text("LATCHING")
    com.put_text("1")
    com.put_text("0")
    com.put_text("0")

    assert pfa2filter.channel_dict[0].overload_clear() is True
    assert com.get_written() == "0:OVLDMODE?"
    assert com.get_written() == "0:OVLDCLEAR?"
    assert com.get_written() == "0:INOVLD?"
    assert com.get_written() == "0:OUTOVLD?"

    com.put_text("CONTINUOUS")
    com.put_text("0")
    com.put_text("0")
    assert pfa2filter.channel_dict[1].overload_clear() is True
    assert com.get_written() == "1:OVLDMODE?"
    assert com.get_written() == "1:INOVLD?"
    assert com.get_written() == "1:OUTOVLD?"

    com.put_text("LATCHING")
    with pytest.raises(Pfa2FilterCommunicationError):
        pfa2filter.channel_dict[0].overload_clear()
